import os, sys, math
import cPickle
import Spectrum
import Peptide

MIN_PEPTIDE_LENGTH = 6
SECONDARY_THRESHOLDS = {'sequest' : 0.10, 'gutentag' : 0.20}
SINGLE_INFO_PER_UALG = True

#---------------------
def totalStats(colls, names):
  string = "Dataset\tReal\tRand\tFDR\n"
  for idx in range(len(colls)):
    coll = colls[idx]
    name = names[idx]
    infos = coll.infos
    real = len(filter(lambda x: x.real == True, infos))
    total = len(infos)
    rand = total - real
    fdr = 1.00
    if real != 0: fdr = float(rand)/real
    string += "%s\t%i\t%i\t%.4f\n" % (name, real, rand, fdr)

  return string

#------------------------------------
# contains the information about a psm generated by a single algorithm
#------------------------------------
class PsmInfo:

  #--------------------
  def __init__(self):
    self.algorithm = ""
    self.groupId = ""
    self.scan = 1
    self.real = False
    self.peptide = ""
    self.charge = 0
    self.primary = 0.0
    self.secondary = 0.0
    
  #--------------------
  def __str__(self):
    string = ""
    string += "%s\t" % self.algorithm
    string += "%s\t" % self.getUid()
    string += "%s\t" % self.peptide
    string += "%s\n" % "n/a"
    return string

  #--------------------
  def toLongString(self):
    string = ""
    string += "%s\t" % self.groupId
    string += "%s\t" % str(self.scan)

    if self.real: string += "real\t"
    else: string += "rand\t"

    string += "%s\t" % self.peptide
    string += "%s\t" % self.charge
    string += "%s\t" % self.primary
    string += "%s\t" % self.secondary
    return string

  #--------------------
  def copy(self):
    copy = PsmInfo()
    copy.algorithm = self.algorithm
    copy.groupId = self.groupId
    copy.scan = self.scan
    copy.real = self.real
    copy.peptide = self.peptide
    copy.charge = self.charge
    copy.primary = self.primary
    try: copy.secondary = self.secondary
    except: copy.secondary =  ""
    return copy

  #------------------------------------
  def parseFromExtendedLine(self, line):
    fields = line.rstrip('\n').split()
    self.algorithm = fields[0]
    self.groupId = fields[1]
    self.scan = fields[2]
    if ('rand' in fields[3]) or ('shuffle' in fields[3]): 
      self.real = False
    else: self.real = True
    self.peptide = fields[4]
    self.charge = int(fields[5])
    self.primary = float(fields[6])
    if len(fields) > 7: self.secondary = float(fields[7])

  #------------------------------------
  def parseFromLine(self, line):
    fields = line.rstrip('\n').split()
    self.groupId = fields[0]
    self.scan = int(fields[1])
    if ('rand' in fields[2]) or ('shuffle' in fields[2]): 
      self.real = False
    else: self.real = True
    self.peptide = fields[3]
    self.charge = int(fields[4])
    self.primary = float(fields[5])
    if len(fields) > 6: self.secondary = float(fields[6])

  #------------------------------------
  def setAlgorithm(self, algorithm):
    self.algorithm = algorithm

  #------------------------------------
  def getUid(self):
    return Spectrum.spectrumUID(self.scan, self.groupId, self.charge)

  #------------------------------------
  def getPsm(self):
    uid = self.getUid()
    psm = (uid, self.peptide)
    return psm

  #------------------------------------
  def getUidAlg(self):
    uid = self.getUid()
    ualg = (uid, self.algorithm)
    return ualg

  #------------------------------------
  def outputPsm(self):
    fields = line.rstrip('\n').split()
    self.groupId = fields[0]
    self.scan = int(fields[1])
    if ('rand' in fields[2]) or ('shuffle' in fields[2]): 
      self.real = False
    else: self.real = True
    self.peptide = fields[3]
    self.charge = int(fields[4])
    self.primary = float(fields[5])
    if len(fields) > 6: self.secondary = float(fields[6])

  #------------------------------------
  # examines both info and determines if self is just a lower scoring version
  # of the otherInfo, from the same scan, groupId and database
  def inferiorInfo(self, otherInfo=None):
    if otherInfo == None: return False
    if (self.charge  == otherInfo.charge  and \
        self.scan    == otherInfo.scan    and \
        self.groupId == otherInfo.groupId and \
        self.real    == otherInfo.real    and \
        self.primary <= otherInfo.primary):
      return True
    else:
      return False

#------------------------------------
# contains the information about multiple psms generated by many algorithm
#------------------------------------
class PsmInfoCollection:
  def __init__(self):
    try:
      del self.infos
      del self.algorithm2infos
      del self.peptide2infos
      del self.peptideCharge2infos
      del self.psm2infos
      del self.uid2infos
      del self.ualg2infos
      del self.charge2infos
      del self.onePepPerAlgPerScan
      del self.onePepPerScan
      del self.realRandsRemoved
      del self.inconsistenciesResolved
    except:
      pass

    self.infos = []
    self.algorithm2infos = {}
    self.peptide2infos = {}
    self.peptideCharge2infos = {}
    self.psm2infos = {}
    self.uid2infos = {}
    self.ualg2infos = {}
    self.charge2infos = {}
    self.onePepPerAlgPerScan = False
    self.onePepPerScan = False
    self.realRandsRemoved = False
    self.inconsistenciesResolved = False

  #------------------------------------
  def copy(self):
    dest = PsmInfoCollection()
    dest.infos = self.infos
    dest.algorithm2infos = self.algorithm2infos
    dest.peptide2infos = self.peptide2infos
    dest.peptideCharge2infos = self.peptideCharge2infos
    dest.psm2infos = self.psm2infos 
    dest.uid2infos = self.uid2infos 
    dest.charge2infos = self.charge2infos
    dest.ualg2infos = self.ualg2infos 
    dest.onePepPerAlgPerScan = self.onePepPerAlgPerScan 
    dest.realRandsRemoved = self.realRandsRemoved 
    try: dest.inconsistenciesResolved = self.inconsistenciesResolved
    except: dest.inconsistenciesResolved = False
    return dest

  #------------------------------------
  def __len__(self):
    return len(self.infos)

  #------------------------------------
  def __str__(self):
    string = ""
    for info in self.infos:
      string += str(info)
    return string 
 
  #------------------------------------
  def toLongString(self):
    string = ""
    for info in self.infos:
      string += info.toLongString() + "\n"
    return string 
 

  #------------------------------------
  def addPsmInfo(self, info):
    ualg = info.getUidAlg()
    if self.ualg2infos.has_key(ualg): 
      # if SINGLE_INFO_PER_UALG == True: return
      self.ualg2infos[ualg].append(info)
    else: self.ualg2infos[ualg] = [ info ] 

    self.infos.append(info)

    if self.charge2infos.has_key(info.charge): 
      self.charge2infos[info.charge].append(info)
    else: self.charge2infos[info.charge] = [ info ] 

    if self.algorithm2infos.has_key(info.algorithm): 
      self.algorithm2infos[info.algorithm].append(info)
    else: self.algorithm2infos[info.algorithm] = [ info ] 

    if self.peptide2infos.has_key(info.peptide): 
      self.peptide2infos[info.peptide].append(info)
    else: self.peptide2infos[info.peptide] = [ info ] 

    if self.peptideCharge2infos.has_key( (info.peptide, info.charge) ): 
      self.peptideCharge2infos[ (info.peptide, info.charge) ].append(info)
    else: self.peptideCharge2infos[ (info.peptide, info.charge) ] = [ info ] 

    if self.psm2infos.has_key(info.getPsm()): 
      self.psm2infos[info.getPsm()].append(info)
    else: self.psm2infos[info.getPsm()] = [ info ] 

    uid = info.getUid()
    if self.uid2infos.has_key(uid): self.uid2infos[uid].append(info)
    else: self.uid2infos[uid] = [ info ] 

  #------------------------------------
  def randomize(self): 
    self.sort( lambda x,y: cmp( random(), random() ) )

  #------------------------------------
  def addFile(self, filename, format='short', parseTopHitOnly=False):
    algorithm = os.path.basename(filename).split('.')[0]
    fh = open(filename)
    lastInfo = None
    for line in fh:
      info = PsmInfo()
      if format == 'extended':
        info.parseFromExtendedLine(line)
      else:
        info.parseFromLine(line)
      info.setAlgorithm(algorithm)
      if (parseTopHitOnly == True) and info.inferiorInfo(lastInfo): 
        lastInfo = info
        continue
      lastInfo = info
      self.addPsmInfo(info)
    fh.close()

  #------------------------------------
  def getScores(self, real=True):
    scores = []
    for info in self.infos:
      if info.real == real:
        scores.append(info.primary)
    return scores

  #------------------------------------
  def getMaxSpScores(self, real=True):
    scores = []
    for info in self.infos:
      if info.real == real:
        scores.append(info.maxSp)
    return scores

  #------------------------------------
  def subsetScansInFile(self, file):
    scans2keep = {}
    fh = open(file)
    for line in fh:
      scan = line.rstrip('\n').split()[0]
      scans2keep[int(scan)] = 1
    keptInfos = []
    for info in self.infos:
      if scans2keep.has_key(int(info.scan)):
        keptInfos.append(info)

    self.__init__() # clear data structures
    for info in keptInfos:
      self.addPsmInfo(info)

  #------------------------------------
  def subsetPepsInFile(self, file):
    peps2keep = {}
    fh = open(file)
    for line in fh:
      peptide = line.rstrip('\n').split()[0]
      peps2keep[peptide] = 1
    keptInfos = []
    for peptide in self.peptide2infos.keys():
      if peps2keep.has_key(peptide):
        for info in self.peptide2infos[peptide]:
          keptInfos.append(info)

    self.__init__() # clear data structures
    for info in keptInfos:
      self.addPsmInfo(info)

  #------------------------------------
  def removeRealRands(self):
    if self.realRandsRemoved == True: return
    newInfos = []
    for peptide in self.peptide2infos.keys():
      hasReal = False
      for info in self.peptide2infos[peptide]: 
        if info.real == True: hasReal = True
      if hasReal == True:
        # remove the rands 
        for info in self.peptide2infos[peptide]: 
          if info.real == True: newInfos.append(info)
      else:
        # take them all
        newInfos += self.peptide2infos[peptide] 

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)
    self.realRandsRemoved = True
   
  #------------------------------------
  def ensureOnePepPerScan(self, normalScore=True):
    if self.onePepPerScan == True: return
    newInfos = []
    for uid in self.uid2infos.keys():
      infos = self.uid2infos[uid]
      if normalScore == True:
        infos.sort(lambda x,y: cmp(y.primary, x.primary))
      else:
        infos.sort(lambda x,y: cmp(x.primary, y.primary))

      # keep the top, regardless of whether rand or real
      newInfos.append(infos[0])

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)
    self.onePepPerScan = True
 
  #------------------------------------
  def ensureOnePepPerAlgPerScan(self):
    if self.onePepPerAlgPerScan == True: return
    newInfos = []
    for ualg in self.ualg2infos.keys():
      algorithm = ualg[1] 
      infos = self.ualg2infos[ualg]
      # if len(infos) > 1: print >>sys.stderr, "Found one"
      if algorithm == 'ommsa':
        infos.sort(lambda x,y: cmp(x.primary, y.primary))
      else:
        infos.sort(lambda x,y: cmp(y.primary, x.primary))

      # keep one rand and one real
      gotRand = gotReal = False
      rand = real = None
      for info in infos:
        if gotRand == True and gotReal == True: break
        if info.real == True: gotReal = True; real = info;
        else: gotRand = True; rand = info;

      if gotRand == True: newInfos.append(rand)
      if gotReal == True: newInfos.append(real)

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)
    self.onePepPerAlgPerScan = True
   
  #------------------------------------
  def minAlgs(self, minAlgs): 
    self.ensureOnePepPerAlgPerScan()
    self.removeRealRands()

    #----
    newInfos = []
    for psm in self.psm2infos.keys():
      if len(self.psm2infos[psm]) < minAlgs: continue
      for info in self.psm2infos[psm]: newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def minLength(self, length):
    #----
    newInfos = []
    for info in self.infos:
      if len(info.peptide) >= length: newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def tryptic(self):
    #----
    newInfos = []
    for info in self.infos:
      if info.peptide[-1] in 'KR': newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def trypticStrict(self):
    #----
    newInfos = []
    for info in self.infos:
      if Peptide.isTryptic(info.peptide, missedCleavagesAllowed=0):
        newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def nonTrypticStrict(self):
    #----
    newInfos = []
    for info in self.infos:
      if not Peptide.isTryptic(info.peptide):
        newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)


  #------------------------------------
  def allowedCharge(self, charge):
    self.ensureOnePepPerAlgPerScan()
    self.removeRealRands()
  
    if type(charge) == type(2):  charge = [charge]
    #----
    newInfos = []
    for info in self.infos:
      if info.charge not in charge: continue
      newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def minScans(self, minScans):
    self.ensureOnePepPerAlgPerScan()
    self.removeRealRands()

    #----
    newInfos = []
    for peptide in self.peptide2infos.keys():
      if len(self.peptide2infos[peptide]) < minScans: continue
      for info in self.peptide2infos[peptide]: newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def filterSecondary(self, secondaryThresholds):
    newInfos = []
    for info in self.infos:
      if secondaryThresholds.has_key(info.algorithm):
        if info.secondary < secondaryThresholds[info.algorithm]: continue
      newInfos.append(info)
    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def resolveInconsistencies(self):
    if self.inconsistenciesResolved == True: return
    newInfos = []
    for uid in self.uid2infos:
      infos = self.uid2infos[uid]
      hasInconsistentPeptides = False
      peptide = infos[0].peptide
      for info in infos[1:]:
        if peptide != info.peptide: hasInconsistentPeptides = True
      if hasInconsistentPeptides == False: newInfos += infos

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)
    self.inconsistenciesResolved = True

  #------------------------------------
  def singleAlgFDR(self, fdr):
    newInfos = []
    for algorithm in self.algorithm2infos:
      infos = self.algorithm2infos[algorithm]
      if algorithm == 'ommsa':
        infos.sort(lambda x,y: cmp( x.primary, y.primary ))
      else:
        infos.sort(lambda x,y: cmp( y.primary, x.primary ))


      for charge in self.charge2infos.keys():
        p = fp = 0.0
        fdrInfos = []
        for info in infos:
          if info.charge != charge: continue
          if info.real == True: p += 1
          else: fp += 1
          if p == 0.0: break
          if (fp/p) > fdr: break
          fdrInfos.append(info)
        
        newInfos += fdrInfos

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def uniquePeptides(self):
    self.ensureOneScanPerPep()
    newInfos = []
    for peptide in self.peptide2infos:
      infos = self.peptide2infos[peptide]
      infos.sort(lambda x,y: cmp(y.primary,x.primary))
      newInfos.append(infos[0])

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def ensureOneScanPerChargePep(self):
    self.ensureOnePepPerAlgPerScan()
    newInfos = []
    for peptideCharge in self.peptideCharge2infos:
      infos = self.peptideCharge2infos[peptideCharge]
      suid2alg = {}
      for info in infos:
        uid = info.getUid()
        if suid2alg.has_key(uid): suid2alg[uid].append(info.algorithm)
        else: suid2alg[uid] = [ info.algorithm ]

      maxAlgUid = ""
      maxAlgs = 0
      for uid in suid2alg.keys():
        if len(suid2alg[uid]) > maxAlgs:
          maxAlgs = len(suid2alg[uid])
          maxAlgUid = uid

      for info in infos:
        if not maxAlgUid == info.getUid(): continue
        else: newInfos += [ info ] 

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)


  #------------------------------------
  def ensureOneScanPerPep(self):
    self.ensureOnePepPerAlgPerScan()
    newInfos = []
    for peptide in self.peptide2infos:
      infos = self.peptide2infos[peptide]
      suid2alg = {}
      for info in infos:
        uid = info.getUid()
        if suid2alg.has_key(uid): suid2alg[uid].append(info.algorithm)
        else: suid2alg[uid] = [ info.algorithm ]

      maxAlgUid = ""
      maxAlgs = 0
      for uid in suid2alg.keys():
        if len(suid2alg[uid]) > maxAlgs:
          maxAlgs = len(suid2alg[uid])
          maxAlgUid = uid

      for info in infos:
        if not maxAlgUid == info.getUid(): continue
        else: newInfos += [ info ] 

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #------------------------------------
  def mergeAlgorithms(self):
    self.ensureOnePepPerAlgPerScan()
    self.resolveInconsistencies()
    newInfos = []
    for uid in self.uid2infos:
      infos = self.uid2infos[uid]
      if len(infos) == 1: 
        newInfos += infos
        continue
      
      algorithms = []
      for info in infos: algorithms.append(info.algorithm)
      algorithms.sort()
      joinedAlgorithms = ','.join(algorithms)

      newInfo = infos[0].copy()
      newInfo.algorithm = joinedAlgorithms

      newInfos += [ newInfo ]

    self.__init__()
    for info in newInfos: self.addPsmInfo(info)

  #-------------------------------------
  def chargeStats(self):
    header =  ['Charge', 'Real', 'Random', 'FDR' ] 
    string = '\t'.join(header) + "\n"
    for charge in self.charge2infos:
      infos = self.charge2infos[charge]
      real = len(filter(lambda x: x.real == True, infos))
      rand = len(infos) - real
      string += "%i\t%i\t%i\t%.4f\n" % (charge, real, rand, float(rand) / real)

    return string

  #-------------------------------------
  def intersectionStats(self):
    algorithms = ['sequest', 'inspect', 'gutentag', 'probid', 'ommsa', \
        'tandem','lutefisk','pepnovo', ]
    header =  algorithms + [ 'Real', 'Random', 'FDR' ] 
    string = '\t'.join(header) + "\n"
    for algorithm in self.algorithm2infos:
      infos = self.algorithm2infos[algorithm]
      algorithmDict = {}
      for algString in algorithm.split(','): algorithmDict[algString] = 1
      for algString in algorithms:
        if algorithmDict.has_key(algString): string +="X\t"
        else: string += "-\t"

      real = len(filter(lambda x: x.real == True, infos))
      rand = len(infos) - real
      string += "%i\t%i\t%.4f\n" % (real, rand, float(rand) / real)

    return string

  #-------------------------------------
  def singleStats(self):
    string = "Algorithm\tReal\tRandom\tFDR\n"
    for algorithm in self.algorithm2infos:
      infos = self.algorithm2infos[algorithm]
      real = len(filter(lambda x: x.real == True, infos))
      rand = len(infos) - real
      string += "%s\t%i\t%i\t%.4f\n" % (algorithm, real, rand, float(rand) / real)
    return string

  #-------------------------------------
  def getFDR(self):
    positives = self.getTP()
    negatives = self.getFP()
    return float(negatives)/positives

  #-------------------------------------
  def getP(self):
    positives = len(filter(lambda x: x.real == True, self.infos))
    return positives

  #-------------------------------------
  def getTP(self):
    positives = len(filter(lambda x: x.real == True, self.infos))
    negatives = self.getFP()
    return positives - negatives

  #-------------------------------------
  def getFP(self):
    negatives = len(filter(lambda x: x.real == False, self.infos))
    return negatives

#------------------------------------
class SqtPsmInfoCollection(PsmInfoCollection):

  #------------------------------------
  def __init__(self):
    PsmInfoCollection.__init__(self)

  #------------------------------------
  def addSqtCollection(self, collection, real=True):
    algorithm = 'sequest'
    for match in collection["matches"]:
      info = PsmInfo()
      info.setAlgorithm(algorithm)
      info.real = bool(real)
      info.groupId = str(collection["fileName"])
      info.scan = int(match["firstScan"])
      info.primary    = info.xcorr  = float(match["xcorr"])
      info.secondary  = info.deltCN = float(match["deltCN"])
      info.rSp = float(match["rSp"])
      info.sp = float(match["sp"])
      info.dM = float(match["dM"])
      info.left = str(match["left"])
      info.peptide = str(match["peptide"])
      info.right = str(match["right"])
      info.charge = int(match["charge"])
      info.maxSp = match["maxSp"]
      self.addPsmInfo(info)


  #------------------------------------
  # Changes primary score!
  def applyPeptideProphet(self):
    for i in self.infos:
      newPrimary = 0.0
      effectiveLength = 0
      if (i.xcorr<=0.0): i.xcorr = 1e-10

      #--- charge 2
      if i.charge == 2:
        if (len(i.peptide) > 15): effectiveLength = 2 * 15
        else: effectiveLength = 2 * len(i.peptide)
        logXcorr = math.log(i.xcorr)/math.log(effectiveLength)
        newPrimary =  8.362  * logXcorr  +  7.386  * i.deltCN + \
                      -0.194 * i.rSp     +  -0.314 * i.dM + \
                      -0.959
      #--- charge 3
      elif i.charge == 3:
        if (len(i.peptide) > 25): effectiveLength = 4 * 25 
        else: effectiveLength = 4 * len(i.peptide)
        logXcorr = math.log(i.xcorr)/math.log(effectiveLength)
        newPrimary = 9.933 * logXcorr + 11.149 * i.deltCN + \
                    -0.201 * i.rSp    + -0.277 * i.dM + \
                    -0.1460
      i.primary = newPrimary

  #------------------------------------
  def applyWashburn(self):
    keptInfos = []
    for info in self.infos:
      if not info.charge in [1,2,3]:
        raise SystemExit, "Strange charge %i" % info.charge

      # deltCN
      if info.secondary < 0.1: continue

      trypticity = Peptide.trypticity(info.peptide, \
          left=info.left, right=info.right, missedCleavagesAllowed=0)

      # tryptic
      if trypticity == 'tryptic':
        if info.charge == 1:
          if info.primary < 1.9: continue
        elif info.charge == 2:
          if info.primary < 2.2: continue
        elif info.charge == 3:
          if info.primary < 3.75: continue
        else:
          continue
      # partially tryptic
      elif trypticity == 'partial':
        if info.charge == 1: 
          continue
        elif info.charge == 2:
          if info.primary < 2.2: continue
        elif info.charge == 3:
          if info.primary < 3.75: continue
        else:
          continue
      # not tryptic
      elif trypticity == 'not':
        if info.charge == 1: 
          continue
        elif info.charge == 2:
          if info.primary < 3.0: continue
        elif info.charge == 3:
          continue
        else:
          continue
      else:
        print >>sys.stderr, "Strange trypticity %s" % trypticity
        continue

      keptInfos.append(info)

    self.__init__()
    for info in keptInfos: self.addPsmInfo(info)


#---
# Begin main
#---
if __name__ == '__main__':
  if len(sys.argv) < 2:
    print >>sys.stderr, "Usage: %s <output-prefix> <peps> <psm-filename-1> ... <psm-filename-n>" % (sys.argv[0])
    sys.exit(1)

  prefix = sys.argv[1]
  fasta  = sys.argv[2]
  psms = sys.argv[3:]

  # preprocessing
  psmInfoColl = PsmInfoCollection()
  for filename in psms: 
    print >>sys.stderr, "Processing %s" % filename
    psmInfoColl.addFile(filename, parseTopHitOnly=True)
    
  print >>sys.stderr, "Min length: length %i" % len(psmInfoColl)
  psmInfoColl.minLength(6)

  print >>sys.stderr, "Removing rands: length %i" % len(psmInfoColl)
  psmInfoColl.removeRealRands()

  print >>sys.stderr, "Secondary filtering: length %i" % len(psmInfoColl)
  psmInfoColl.filterSecondary(SECONDARY_THRESHOLDS)

  #---- Remove for multi
  # print >>sys.stderr, "Charge +2 filtering: length %i" % len(psmInfoColl)
  # psmInfoColl.allowedCharge(2)

  print >>sys.stderr, "Ensuring one pep per alg per scan: length %i" % len(psmInfoColl)
  psmInfoColl.ensureOnePepPerAlgPerScan()

  print >>sys.stderr, "Tryptic: length %i" % len(psmInfoColl)
  psmInfoColl.trypticStrict()

  # print >>sys.stderr, "E coli tryptic peps: length %i" % len(psmInfoColl)
  # psmInfoColl.subsetPepsInFile(fasta)

  print >>sys.stderr, "Dumping: length %i" % len(psmInfoColl)

  cPickle.dump(psmInfoColl, open(prefix + ".psminfocoll", "w"))

  # do the single alg processing
  # psmInfoColl.singleAlgFDR(FDR)
  # fh = open(prefix + ".singlestats", 'w')
  # print >>fh, psmInfoColl.singleStats()
  # fh.close()
  # psmInfoColl.mergeAlgorithms()
  # fh = open(prefix + ".mergestats", "w")
  # print psmInfoColl.singleStats()
  # fh.close()

  # psmInfoColl.subsetPepsInFile(fasta)
  # psmInfoColl.minAlgs(2)
  # psmInfoColl.minScans(2)
  # psmInfoColl.randomize()
  
  # fh = open(output, 'w')
  # print >>fh, psmInfoColl,
  # fh.close()

  



