import math
import sys
from math import *

def make_gaussian_xy(mean, stddev, minimum, maximum):
  """Given a mean and stddev (mu and sigma) of a Gaussian distribution, make 
  a series of x and y values to be used to plot the Gaussian, where the
  x values between minimum and maximum"""
  width  =  maximum - minimum
  numberXs = 1000
  xs = [  minimum + width*i/numberXs for i in range(numberXs) ]
  ys = []
  for x in xs:
    numerator = exp( -pow(x-mean, 2.0)/(2*pow(stddev,2.0)) )
    denominator = stddev * pow(2*math.pi, 0.5)
    ys.append(numerator/denominator)

  return xs, ys

##############################################################################
def make_gaussians_xy(means, vars, weights, minimum, maximum):
  sumys = []
  stddevs = [ pow(i, 0.5) for i in vars]
  for idx in range(len(means)):
    xs, ys = make_gaussian_xy(means[idx], stddevs[idx], minimum, maximum)
    ys = [ i * weights[idx] for i in ys ]
    if len(sumys) != 0:
      sumys = [ ys[i] + sumys[i] for i in range(len(ys))]
    else:
      sumys = ys
    # pylab.plot(xs,ys)
    # pylab.hold(True)

  return xs, sumys


##############################################################################
def make_evd_xy(mu, lamda, minimum, maximum):
  width  =  maximum - minimumm
  xs = [  minimum + width*i/100.0 for i in range(100) ]
  ys = []
  for x in xs:
    temp = -lamda*(x-mu) 
    y = lamda * exp( temp - exp(temp) )
    ys.append(y)

  return xs, ys

##############################################################################
# Fit an extreme value distribution to a set of values.
# Return the parameters mu and lambda.
def fit_evd(values, starting_L, epsilon):
  """
  This function finds the location parameter, mu, and scale parameter, 1/L, 
  that maximize the log likelihood of the data given an extreme value 
  distribution.  It finds the parameters by using Newton-Raphson to find 
  the zero of the constraint function.  The zero of the constraint function 
  corresponds to the scale parameter giving the maximum log likelihood for the 
  data.

  The parameter values contains the list of the data values.
  The parameter starting_L contains a staring guess for L.
  The parameter contains the tolerence for determining convergence.

  Returns the values of mu and L that maximize the log likelihood.
  Throws an exception if Newton-Raphson fails to converge.
  """
  # Use Newton-Raphson to find best value for scale parameter.
  L = starting_L
  max_iterations = 10000
  for i in range(max_iterations):
    (f, fprime) = constraint_function(values, L)
    if abs(f) < epsilon:
      break
    else:
      L = L - f / fprime 
  if i >= max_iterations:
    raise Exception('Root finding failed to converge.')

  # Calculate best value of position parameter from best value of 
  # scale parameter.
  exponentials = map(lambda x : math.exp(-L * x), values)
  mu = -1.0 / L * math.log(1.0 / len(values) * sum(exponentials))

  return (mu, L)

##############################################################################

def censored_fit_evd(values, threshold, starting_L, epsilon):
  """
  This function finds the location parameter, mu, and scale parameter, 1/L, 
  that maximize the log likelihood of the data given an extreme value 
  distribution.  It finds the parameters by using Newton-Raphson to find 
  the zero of the constraint function.  The zero of the constraint function 
  corresponds to the scale parameter giving the maximum log likelihood for the 
  data.

  Data values less then the value of threshold will be censored. That is, if
  the number of values less then the threshold is z, the likelihood of the data 
  will depend only on the probability of finding z values less then the 
  threshold, not on the probability of the individual values.

  The parameter values contains the list of the data values.
  The parameter threshold sets the value under which data
  will be censored.
  The parameter starting_L contains a staring guess for L.
  The parameter contains the tolerence for determining convergence.

  Returns the values of mu and L that maximize the log likelihood.
  Throws an exception if Newton-Raphson fails to converge.
  """
  # Use Newton-Raphson to find best value for scale parameter.
  L = starting_L
  max_iterations = 10000
  uncensored_values = [ x for x in values if x > threshold ]
  num_censored = len(values) - len(uncensored_values)
  for i in range(max_iterations):
    (f, fprime) = censored_constraint_function(uncensored_values, 
      threshold, num_censored, L)
    if abs(f) < epsilon:
      break
    else:
      L = L - f / fprime 
  if i >= max_iterations:
    raise Exception('Root finding failed to converge.')

  # Calculate best value of position parameter from best value of 
  # scale parameter.
  exponentials = map(lambda x : math.exp(-L * x), uncensored_values)
  mu = -1.0 / L * log(1.0 / len(uncensored_values) * \
                      (num_censored * math.exp(-L * threshold) + sum(exponentials)))

  return (mu, L)

def constraint_function(values, L):
  """
  This function is a transformation of the partial derivatives of
  the log likelihood of the data given an extreme value distribution
  with location parameter mu and scale parameter 1/L. The transformation 
  has eliminated the explicit dependence on the location parameter, mu, 
  leaving only the scale parameter, 1/L.

  The zero crossing of this function will correspond to the maximum of the 
  log likelihood for the data.

  See equations 10 and 11 of "Maximum Likelihood fitting of extreme value 
  distributions".

  The parameter values contains a list of the data values.
  The parameter L is the reciprocal of the scale parameters.

  Returns the value of the function and its derivative.
  """
  exponentials = map(lambda x : math.exp(-L * x), values)
  numerator = sum(map(lambda x, y: x * y, values, exponentials))
  denominator = sum(exponentials)
  function = (1.0 / float(L)) - (sum(values) / float(len(values))) \
    + (numerator / denominator)
  second_numerator \
    = sum(map(lambda x, y: (x ** 2) * y, values, exponentials))
  derivative = ((numerator ** 2) / (denominator ** 2)) \
    - ((second_numerator / denominator)) - (1.0 / (L ** 2))

  return (function, derivative)

def censored_constraint_function(values, threshold, num_censored, L):
  """
  This function is a transformation of the partial derivatives of
  the log likelihood of the data given an extreme value distribution
  with location parameter mu and scale parameter 1/L. The transformation 
  has eliminated the explicit dependence on the location parameter, mu, 
  leaving only the scale parameter, 1/L.

  The zero crossing of this function will correspond to the maximum of the 
  log likelihood for the data.

  See equations 10 and 11 of "Maximum Likelihood fitting of extreme value 
  distributions".

  The parameter values contains a list of the data values.
  The parameter threshold contains the value below which data were censored.
  The parameter num_censored is the count of how many values were removed 
  from the data set.
  The parameter L is the reciprocal of the scale parameters.

  Returns the value of the function and its derivative.
  """
  exponentials = map(lambda x : math.exp(-L * x), values)
  numerator = (num_censored * threshold * math.exp(-L * threshold)) \
    + sum(map(lambda x, y: x * y, values, exponentials))
  denominator = num_censored * math.exp(-L * threshold) + sum(exponentials)
  function = (1.0 / float(L)) - (sum(values) / float(len(values))) \
    + (numerator / denominator)
  second_numerator = num_censored * (threshold ** 2) * math.exp(-L * threshold) \
    + sum(map(lambda x, y: (x ** 2) * y, values, exponentials))
  derivative = (numerator / denominator) ** 2 \
    - (second_numerator / denominator) - (1.0 / (L ** 2))

  return (function, derivative)

##############################################################################
# Compute a p-value for a given score w.r.t. an EVD with the given parameters.
# Function: ExtremeValueP()
# 
# Purpose:  Calculate P(S>x) according to an extreme
#           value distribution, given x and the parameters
#           of the distribution (characteristic
#           value mu, decay constant lambda).
#           
#           This function is exquisitely prone to
#           floating point exceptions if it isn't coded
#           carefully.
#           
# Args:     x      = score
#           mu     = characteristic value of extreme value distribution
#           lambda = decay constant of extreme value distribution
#           
# Return:   P(S>x)
#
def compute_evd_pvalue(score, evd_mu, evd_lambda):

  # These constants are hardware dependent.
  # These values should be good for double precision floating point
  # numbers compatible with the IEEE 754 standard.
  DBL_EPSILON = 2.2204460492503131e-16
  DBL_MAX_10_EXP = 308

  # avoid exceptions near P=1.0
  if ((evd_lambda * (score - evd_mu))
      <= -1. * math.log(-1. * math.log(DBL_EPSILON))):
    return 1.0
  # avoid underflow fp exceptions near P=0.0
  if ((evd_lambda * (score - evd_mu)) >= 2.3 * DBL_MAX_10_EXP):
    return 0.0
  # a roundoff issue arises; use 1 - e^-x --> x for small x
  y = math.exp(-1.0 * evd_lambda * (score - evd_mu))
  if (y < 1e-7):
    return y
  else:
    return (1.0 - math.exp(-1.0 * y))

def fit_gaussian(scores):
  """Given a list of scores, calculate the maximum likelihood mean and
  stddev (mu and sigma) of the underlying Gaussian distribution"""
  mean = sum(scores)/len(scores)
  variance = sum(map(lambda x: pow(x - mean, 2.0), scores))/len(scores)
  stddev = pow(variance, 0.5)
  return mean, stddev

#-------------------------------
# Computes a zscore (which can be converted to a p-value) from a mean and stddev
#-------------------------------
def compute_gaussian_zscore(score, mean, stddev):
  return (score - mean)/variance

if __name__ == '__main__':
  import random
  values = []
  for i in range(100):
    valuesToMax = [ random.random() for i in range(1000) ]
    values.append(max(valuesToMax))

  
  mu, lamda = fit_evd(values, 0.90, 0.01)
  mu, lamda = censored_fit_evd(values, 0.99, 0.90, 0.01)
  print str(compute_evd_pvalue(1.10, mu, lamda))
    
##############################################################################
def make_weibull_y(eta, beta, xs):
  ys = []
  first_factor = beta / eta

  new_xs = []
  for x in xs:
    second_factor = pow(x/eta, beta - 1.0)
    third_factor  = pow(math.e, - pow(x/eta, beta))
    y = first_factor * second_factor * third_factor
    ys.append(y)

  return ys 


##############################################################################
def make_weibull_xy(eta, beta, minimum, maximum, normed=False, shift=0.0):
  width  =  maximum - minimum
  numberXs = 1000
  xs = [  0.0001 + width*i/numberXs for i in range(numberXs) ]
  ys = []
  first_factor = beta / eta
  if normed == True:
    correction = pow(math.e, - pow(minimum/eta, beta))

  xs.sort(lambda x,y: cmp(y,x))
  new_xs = []
  for x in xs:
    shifted_x = x + shift
    if (shifted_x <= 0.0):
      print >>sys.stderr, "Negative shifted %.3f" % x
      break
    new_xs.append(x)
    second_factor = pow(shifted_x/eta, beta - 1.0)
    third_factor  = pow(math.e, - pow(shifted_x/eta, beta))
    y = first_factor * second_factor * third_factor
    if normed == True:
      y /= correction
    ys.append(y)

  # new_ys = map(lambda y: y * len(new_xs) / len(xs), ys)

  return new_xs, ys 






