#-------------------
# $Version$
#-------------------
# $Author: aklammer $
#-------------------
 
#-------------------
import cPickle, re, sys, os
import readLPs
import Spectrum
import Mass
from hmmParams import *
import GMTKmodel

#--------------------------------------
class GMTKprobs:
#--------------------------------------
  def __init__(self):
    self.nterm = {}
    self.cterm = {}
    for amino in Mass.CanonicalAminos:
      self.nterm[amino] = None
      self.cterm[amino] = None
    self.region = [0] * NUM_REGIONS
    self.mixtures = []


  #--------------------------------------
  def parseProbs(self, model, paramPrefix, ionType):
    """ <model> is the model for which probabilities will be parsed
        <paramPrefix> is the prefix of the parameter files to parse
        <ionType> is the kind of ion for which to parse"""

    modelDir = "%s/model-%s/params" % (MODEL_DIR, model.name)
    paramFile = modelDir + os.sep + paramPrefix + "," + ionType + ",pos.params"

    for aminoIdx in range(len(Mass.CanonicalAminos)): 
      #----- above assumes consistency with Mass module
      #----- parse the amino N-term
      aminoString = Mass.CanonicalAminos[aminoIdx]
      self.nterm[aminoString] = \
          readLPs.getParamInfo(['', paramFile, 'iNaaId_%i' % aminoIdx])
      #----- parse the amino C-term
      self.cterm[aminoString] = \
          readLPs.getParamInfo(['', paramFile, 'iCaaId_%i' % aminoIdx])

    #----- parse the region distribution
    for regionIdx in range(NUM_REGIONS):
      self.region[regionIdx] = \
          readLPs.getParamInfo(['', paramFile, 'iRegion_%i' % regionIdx])
    #----- parse the N-term, C-term, region weights
    self.parseMixtureWeights(paramFile)
     
  #--------------------------------------
  def parseMixtureWeights(self, paramFile):
  #--------------------------------------
    fh = open(paramFile)
    inregion = False
    regionLines = []
    for line in fh:
      if line.startswith("iMix_iRegion"):
        inregion = True
      if inregion == True and line.startswith("\n"):
        break
      if inregion == True:
        regionLines.append(line)
    fh.close()
    for idx in range(3,len(regionLines)):
      self.mixtures.append(map(float, regionLines[idx].rstrip('\n').split()))
    # mixtures is first column: N-term C-term then Region
    # print self.mixtures

  #--------------------------------------
  def getValue(self, value='modeX', term='N', amino='P'):
    if term == 'N':
      return self.nterm[amino][value]
    elif term == 'C':
      return self.cterm[amino][value]
    else:
      raise SystemExit, "Bad term %s" % term

  #--------------------------------------
  def getDist(self, term='N', amino='P'):
    if term == 'N':
      return self.nterm[amino]['points']
    elif term == 'C':
      return self.cterm[amino]['points']
    else:
      raise SystemExit, "Bad term %s" % term

  #--------------------------------------
  def calcDist(self, nterm=None, cterm=None, region=None):
    #----- calculate the full distribution from the N-term C-term region
    #----- should we cache?
    ntermDist  = self.nterm[nterm]['points']
    ctermDist  = self.cterm[cterm]['points']
    regionDist = self.region[region]['points']
    finalDist = []
    for idx in range(len(regionDist)):
      finalDist.append(\
          ntermDist[idx] * self.mixtures[region][0] + \
          ctermDist[idx] * self.mixtures[region][1] + \
          regionDist[idx] * self.mixtures[region][2])
    finalDist = [ i/3.0 for i in finalDist ]
    return finalDist

#--------------------------------------
def indexOfMax(array):
  max = 0.0
  maxIdx = None
  for idx in range(len(array)):
    if array[idx] > max:
      max = array[idx]
      maxIdx = idx
  return maxIdx

#--------------------------------------
def mean(array):
  sum = 0.0
  for idx in range(len(array)):
    sum += float(idx) * array[idx]
  return sum/len(array)/len(array)

#--------------------------------------
class GMTKprobCollection:
#--------------------------------------
  def __init__(self, model=None):
    self.ionTypeProbs = {}
    self.model = model
    for ionType in self.model.fileTypes:
      self.ionTypeProbs[ionType] = None
    
    self.cachedDist = {}
    self.cachedProb = {}

  #--------------------------------------
  def parse(self, paramPrefix, ionTypes=None):

    if ionTypes == None:
      ionTypes = self.model.fileTypes

    #----- parse a GMTK prob for each ionType
    for ionType in ionTypes:
      if ionType == 'p': continue
      myDist = GMTKprobs()
      myDist.parseProbs(self.model, paramPrefix, ionType)
      self.ionTypeProbs[ionType] = myDist

  #--------------------------------------
  def calcDist(self, nterm, cterm, region, ionType):
    key = ','.join(map(str, [nterm,cterm,region,ionType]))
    if not self.cachedDist.has_key(key): 
      self.cachedDist[key] = \
          self.ionTypeProbs[ionType].calcDist(nterm, cterm, region)
    return self.cachedDist[key]

  #--------------------------------------
  def getDist(self, term='N', amino='P', ionType='y'):
    return self.ionTypeProbs[ionType].getDist(term=term, amino=amino)

  #--------------------------------------
  def getValue(self, value='modeX', term='N', amino='P', ionType='y'):
    return \
      self.ionTypeProbs[ionType].getValue(value=value, term=term, amino=amino)

  #--------------------------------------
  def calcProb(self, nterm, cterm, region, ionType, type='mode'):
    key = ','.join(map(str, [nterm,cterm,region,ionType,type]))
    if not self.cachedProb.has_key(key): 
      dist = self.calcDist(nterm, cterm, region, ionType)
      if type == 'mode':
        self.cachedProb[key] = indexOfMax(dist) / float(len(dist))
      if type == 'mean':
        self.cachedProb[key] = mean(dist)
      # remember not to cache when we are sampling
    return self.cachedProb[key]

#-------
# Main
#-------
if __name__ == '__main__':

  import pylab
  if len(sys.argv) < 2:
    raise SystemExit, "Usage: %s <model-name> <param-prefix>" % (sys.argv[0])

  modelName, paramPrefix = sys.argv[1], sys.argv[2]
  model = GMTKmodel.gmtkModelFactory(modelName, paramPrefix=paramPrefix)
  probs = GMTKprobCollection(model)
  # TODO transofrm modelname to model
  probs.parse(paramPrefix)
  ys = probs.calcDist('N','P',1, 'y')
  xs = range(len(ys))
  pylab.plot(xs,ys)
  pylab.show()
  #import cPickle
  #psm = cPickle.load(open(psmFile))

