#!/usr/bin/python
#$Revision: 1.19 $
import getopt,sys,re,copy,os.path
import Hit
from Match import *


class HitCollection:
	"A collection of protein hits."
	# To be implemented in the sub-classes
	def nextHit(self):
		pass

class DtaSelectHitCollection(HitCollection):
	"Parses out the information in a DTASelect file."

	# Constructor takes a file as input.
	def __init__(self, filename):
		self.data = {}
		self.data["filename"] = filename
		self.data["file"] = open(filename, 'r')
		self.data["cachedLine"] = ""
		self.data["headerLines"] = []
		self.data["footerLines"] = []
		# should be moved to the parent class
		self.data["hits"] = []

	# TODO need to add __del__ method and __str__ method.

	# at some point have ability to say "protein" and retrieve a hit
	# or "peptide" and retrieve list of hits
	def __getitem__(self, key):
		return self.data[key]

	def __setitem__(self, key, item):
		self.data[key] = item		

	def shortname(self):
		return os.path.basename(os.path.dirname( \
			self.data["filename"]).replace("/$",""))
	
#	def parseFirstLine(self, l):
#		results = re.split("\t", l)
#		return results[1], results[2], results[3]

	def nextHit(self):
		"Returns the next Hit in the file."
		# read in header
		if self["cachedLine"] == "":
			l = self["file"].readline()
			if l == "": return None
			while l[0:6] != "Unique":
				self["headerLines"].append(l)
				l = self["file"].readline()
				if l == "": return None
			self["headerLines"].append(l)
			self["cachedLine"] = self["file"].readline()
		
		lines = [self["cachedLine"]]

		l = self["file"].readline()
		if l == "": return None
	  	
		# read in peptide match
		# these are *unique* matches. the redundancy of the match is kept in
		# each match
		matches = []
		while l[0] == "*" or ((l[0] == "\t") and not re.search("^\tProt", l)):
			lines.append(l)

			if l[0] == "*":
				# FIX use default values for now. if we need to pull it out we will
				match = SqtMatch()
				matches.append(match)	
				list = l.split("\t")
				match["redundancy"] = int(list[10])
				match["id"] = list[1]
				match["peptide"] = list[11].replace("\n","").split('.')[1]
				id_list = list[1].split(".")
				match["first scan"] = int(id_list[1])
				match["last scan"] = id_list[2]
				match["charge"] = int(id_list[3])
				match["allLines"].append(l)


			l = self["file"].readline()
			if l == "": break
		
		self["cachedLine"] = l

		# read in footer
		if re.search("^\tProt", l):
			while l != "":
				self["footerLines"].append(l)
				l = self["file"].readline()
	
		hit = Hit.DtaSelectHit()
		hit["matches"] = matches

		self.parseHitInfo(hit, lines) 
		return hit 
		
	def parseHitInfo(self,hit,lines):
		"Parses protein id and type of protein from the protein descriptor line."

		hit["allLines"] = copy.deepcopy(lines)
		split_line = lines[0].split("\t")
		hit["protein"] = split_line[0] 
		hit["num peptides"] = split_line[1] 
		hit["num spectra"] = split_line[2] 
		if re.search("contam", hit["protein"]):
			hit["type"] = "contam"
		elif re.search("rand", hit["protein"]):
			hit["type"] = "random"
		else: 
			hit["type"] = "real"

		# we don't have any confidence metric yet, but we will someday
		# hit["confidence_metric"] = 


	def allHits(self):
		"Returns a list of all the Hits in this collection."

		if len(self["hits"])==0:
			# we must not have read in the hits yet
			hit = self.nextHit()
			while hit != None:
				self["hits"].append(hit)
				hit = self.nextHit()

		return self["hits"]


	def subtractHits(self, dtaHitCollection):
		"Returns a list of hits that are in self but not in the dtaHitColl."

		subbed_hits = [] 


		his_hits = dtaHitCollection.allHits()
		his_prots_dict = {}
		for hit in his_hits:
			his_prots_dict[hit["protein"]] = 1 

		for hit in self.allHits():
			if not his_prots_dict.has_key(hit["protein"]):
				subbed_hits.append(hit)

		return subbed_hits


	def calcFP(self):
		"Estimates the number of false positive protein hits."

		hits = self.allHits()
		return len(filter(lambda x: x["type"] != "real", hits))


	def calcFPmatches(self):
		"Estimates the number of false positive peptide-spectrum matches."

		hits = self.allHits()
		fpHits = filter(lambda x: x["type"] != "real", hits)
		match_sum = 0 
		for hit in fpHits:
			for match in hit["matches"]:
				match_sum = match_sum + match["redundancy"]

		return match_sum


	def calcTP(self):
		"Estimates the number of true positive protein hits."
	
		hits = self.allHits()
		fpHits = self.calcFP()

		return len(hits)  - 2 * fpHits


	def calcTPmatches(self):
		"Estimates the number of true positive peptide-spectrum matches."

		hits = self.allHits()
		fpMatches = self.calcFPmatches()

		match_sum = 0 
		for hit in hits:
			for match in hit["matches"]:
				match_sum = match_sum + match["redundancy"]

		return (match_sum - 2 * fpMatches)


	def pids(self):
		hits = self.allHits()
		print len(hits)
		realHits = filter(lambda x: x["type"] == "real", hits)
		pids = []
		for hit in realHits:
			pids.append(hit["protein"].split("|")[1])
		return pids
	


if __name__ == '__main__':
	if len(sys.argv) < 3:
		print >>sys.stderr, "Usage: %s <marray-intensity2pid> <DTASelect filename 1> ... < DTASelect filename n>"  % (sys.argv[0])
		sys.exit(1)

	gi2intensity = {}
	intensity_file = open(sys.argv[1], 'r')
	for line in intensity_file:
		intensity, gi = line.split()
		gi2intensity[gi] = intensity

	my_dta = []
	for i in range(2,len(sys.argv)):
		my_dta.append(DtaSelectHitCollection(sys.argv[i]))

	gi2spectra = {}
	for i in range(len(my_dta)):
		hits = my_dta[i].allHits()
		tpHits = filter(lambda x: x["type"] == "real", hits)
		for hit in tpHits:
			gi, spectra = hit["gi"], hit["num spectra"]
			if gi2spectra.has_key(gi):
				gi2spectra[gi]+=int(spectra)
			else:
				gi2spectra[gi]=int(spectra)

	
	# print out the spectrum number vs. the intensity
	for gi in gi2spectra:
		if gi2intensity.has_key(gi):
			print gi, "\t", gi2spectra[gi], "\t", gi2intensity[gi]
		else:
			print >>sys.stderr, "Missing intensity for gi %s" % gi

#	hit = dta.nextHit()
#	for i in dta["headerLines"]:
#		print i, 
#
#	list = []
#	while hit != None:
#		list.append(hit)
#		hit = dta.nextHit()
#		
#	for hit in list:	
#		for i in hit["allLines"]:
#			print i,
#
#	for i in dta["footerLines"]:
#		print i,
#	
#
