#!/usr/bin/python
#--------------------
# $Revision: 1.40 $
#--------------------
import sys, re, math, random, cPickle
from compile_aa_stats import *
import Mass, HmmData
# from Utils import *
# from Enum import Enum
# from Mass import avg_mass, tol_mass, mono_mass
from MatchCollection import *
try:
	from MobileProtonModel import mobilizeProtons
except:
	print >>sys.stderr, "Could not import MobileProtonModel!"

import GMTKprobs

model2legend = {\
  "MPROTON-BY": "MPROTON-BY", \
  "PAIR-ADJ": "PAIR-ADJ",\
  "UNI": "UNI"}

aas = ['A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R',\
    'S','T','V','W','Y']

DEFAULT_INTENSITY = 1
MIN_MZ = 400
MAX_MZ = 2400

#-------------------------------------------------------------------------------
class Model:
  """Contains information for setting the intensities of a Spectrum"""

  def __init__(self):
    self.shortname = "None"

  def __getitem__(self, key):
    return self.data[key]

  def __setitem__(self, key, item):
    self.data[key] = item    

  def __str__(self):
    return "Model of type %s\n" % self.type

  def addIntensityInfo(self, spectrum):
    """Checks arguments"""
    # TODO possibly check to see if spectrum has the necessary info
    # for that particular model
    # TODO add an argument checker to the base class
    pass

  def train(self, psmMatchCollection):
    for match in psmMatchCollection.getMatches():
      match.spectrum.sumNormalizeIntensity()

#-------------------------------------------------------------------------------
class RandomModel(Model):
  """Random model"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "RAND"

  def train(self, psmMatchCollection):
    # Requires no training. Sets rand num of peaks to rand heights
    pass
  
  def addIntensityInfo(self, spectrum):
    # count the number of b and y ions
    peakNum = 0
    for peak in spectrum.getPeaks():
      if peak.isBpeak() or peak.isYpeak(): peakNum += 1
    for peakCount in range(peakNum):
      # get a random mz
      mz = (MAX_MZ-MIN_MZ)*random.random() + MIN_MZ
      i = random.random()
      r = RichPeak(mz=mz,z=1,i=i)
      spectrum.addPeak(r)

#-------------------------------------------------------------------------------
class RandomHeightModel(Model):
  """Random Height model"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "RAND-HT"

  def train(self, psmMatchCollection):
    # This function requires no training. Sets b-y to random heights.
    pass
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      if peak.isBpeak() or peak.isYpeak(): peak.i = random.random()

#-------------------------------------------------------------------------------
class UniformModel(Model):
  """Uniform intensity model"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "UNI" 

  def train(self, psmMatchCollection):
    # This function requires no training. Set b-y to uniform heights.
    pass
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      if peak.isBpeak() or peak.isYpeak(): peak.i = DEFAULT_INTENSITY

#-------------------------------------------------------------------------------
class RealByUniformModel(Model):
  """Uniform real B-Y ions"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "REAL-BY-UNI" 

  def train(self, psmMatchCollection):
    # This function requires no training. 
    pass
  
  def addIntensityInfo(self, spectrum):
    originalObserved = spectrum["source spectrum"]
    spectrum.addIntensitiesFromSpectrum(originalObserved)
    for peak in spectrum.getPeaks():
      if peak.isBpeak():
        peak.i  = (peak.i + peak.partnerPeak.i) / 2.0
        peak.partnerPeak.i = peak.i

#-------------------------------------------------------------------------------
class RealByModel(Model):
  """Real B-Y ions"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "REAL-BY" 

  def train(self, psmMatchCollection):
    # This function requires no training. 
    pass
  
  def addIntensityInfo(self, spectrum):
    originalObserved = spectrum["source spectrum"]
    originalObserved.parsePeaks()
    spectrum.addIntensitiesFromSpectrum(originalObserved)

#--
TEMP_DIR = "mproton"

#-------------------------------------------------------------------------------
class MobileProtonByModel(Model):
  """Mobile Proton B-Y ions"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "MPROTON" 

  def train(self, psmMatchCollection):
    # This function requires no training. 
    pass
  
  def addIntensityInfo(self, spectrum):
    if not os.path.isdir(TEMP_DIR): os.mkdir(TEMP_DIR)
    if int(spectrum["charge"]) == int(3): return
    mprotonFilename = mobilizeProtons(spectrum["charge"], spectrum.peptide, TEMP_DIR)
    if mprotonFilename == None:
      print >>sys.stderr, "Could not get a spectrum file for %i %s" % \
          (spectrum["charge"], spectrum.peptide)
      # TODO change to an exception
      # raise SystemExit


    # parse the Mobile proton output file to a spectrum object
    mproton = DtaSpectrum(mprotonFilename, parsePeaks=True)
    
    spectrum.addIntensitiesFromSpectrum(mproton)

#-------------------------------------------------------------------------------
class MobileProtonModel(Model):
  """Mobile Proton B-Y ions"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "MPROTON" 

  def train(self, psmMatchCollection):
    # This function requires no training. 
    pass
  
  def addIntensityInfo(self, spectrum):
    if not os.path.isdir(TEMP_DIR): os.mkdir(TEMP_DIR)
    if int(spectrum["charge"]) == int(3): return
    mprotonFilename = mobilizeProtons(spectrum["charge"], spectrum.peptide, TEMP_DIR)
    if mprotonFilename == None:
      print >>sys.stderr, "Could not get a spectrum file for %i %s" % \
          (spectrum["charge"], spectrum.peptide)
      # TODO change to an exception
      raise SystemExit

    # parse the Mobile proton output file to a spectrum object
    mproton = DtaSpectrum(mprotonFilename, parsePeaks=True)

    # add ALL peaks to the spectrum
    for peak in mproton.getPeaks():
      spectrum.addPeak(RichPeak(mz=peak[0], i=peak[1]))


#-------------------------------------------------------------------------------
# Divide spectrum into fifteen bins relative to the precursor m/z (not mass)
# Starting at 0.10 ensures that peaks that fall near integer multiples 
# of the precursor ion will fall into the same bin
#-------------------------------------------------------------------------------
MIN_PRECURSOR_RATIO = 0.10
BIN_SIZE = 0.20
NUMBER_MZ_BINS = 20
# MAX = 4.10
#-------------------------------------------------------------------------------
class PrecursorMzModel(Model):
  """Model that corrects for m/z bias relative to precursor intensity"""

  def __init__(self):
    Model.__init__(self)
    self.shortname = "ADJ" 
    self.binSums = range(NUMBER_MZ_BINS)
    # init binSums
    for idx in range(NUMBER_MZ_BINS):
      self.binSums[idx] = 0.0

  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    numberMatches = 0
    for match in psmMatchCollection.getMatches():
      numberMatches += 1
      precursorMz = match.spectrum["precursor mz"]
      for peak in match.spectrum.getPeaks():
        mz, i  = peak[0], peak[1]
        mzBin = int( ((mz/precursorMz) - MIN_PRECURSOR_RATIO) / BIN_SIZE )
        assert mzBin < NUMBER_MZ_BINS
        self.binSums[mzBin] += i
    for idx in range(NUMBER_MZ_BINS):
      self.binSums[idx] /= numberMatches

  def getBinAverage(self, mzBin): 
    return self.binSums[mzBin]

  # set each peak to the expected intensity for its mz bin
  def addIntensityInfo(self, spectrum):
    precursorMz = spectrum["precursor mz"]
    for peak in spectrum.getBYpeaks():
      # if peak.isBpeak() or peak.isYpeak(): peak.i = DEFAULT_INTENSITY
      mzBin = int( ((peak.mz/precursorMz) - MIN_PRECURSOR_RATIO) / BIN_SIZE )
      peak.i = self.binSums[mzBin]

#-------------------------------------------------------------------------------
class AdjusterModel(Model):
  """Parent class of most complex models. Has a HASA relationship
  with PrecursorMzModel. Not used to predict intensities
  itself, but to provide common functionality necessary to train
  most complex models. Training an AdjusterModel model on a
  psmMatchCollection will train a PrecursorMzModel, and then use
  that model to normalize the intensities of peaks within m/z bins
  so that the peaks in each bin to sum to an equal value ACROSS
  ALL SPECTRA. Thus, training carries out a normalization similar to 
  that used by SEQUEST, but maintains important peak intensity
  information."""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "ADJ" 
    self.precursorMzModel = precursorMzModel

  def train(self, psmMatchCollection):
    """WARNING: This method changes the psmMatchCollection"""
    Model.train(self, psmMatchCollection)
    if self.precursorMzModel == None: 
      self.precursorMzModel = PrecursorMzModel()
      self.precursorMzModel.train(psmMatchCollection)
    self.adjustMatchCollectionIntensities(psmMatchCollection)
  
  def adjustMatchCollectionIntensities(self, psmMatchCollection):
    for match in psmMatchCollection.getMatches():
      precursorMz = match.spectrum["precursor mz"]
      peaks = match.spectrum.getPeaks()
      for idx in range(len(peaks)):
        mz = peaks[idx][0]
        i  = peaks[idx][1]
        mzBin = int( ((mz/precursorMz) - MIN_PRECURSOR_RATIO) / BIN_SIZE )
        # notice division below
        newIntensity = i / self.precursorMzModel.getBinAverage(mzBin)
        newPeak = (mz, newIntensity)
        peaks[idx] = newPeak

  def addIntensityInfo(self, spectrum):
    raise SystemExit, "AdjusterModel.addIntensityInfo is not a valid method!"

  # reduce each peak to the expected intensity for its mz bin
  def adjustIntensityInfo(self, spectrum):
    precursorMz = spectrum["precursor mz"]
    peaks = spectrum.getPeaks()
    for idx in range(len(peaks)):
        mz = peaks[idx].mz
        i  = peaks[idx].i
        mzBin = int( ((mz/precursorMz) - MIN_PRECURSOR_RATIO) / BIN_SIZE )
        # notice multiplication below
        newIntensity = i * self.precursorMzModel.getBinAverage(mzBin)
        peaks[idx].i = newIntensity

# FIX add argument checking to these to determine whether Ms2Collection
# or RichSpectrum
# Also are intensities in RichSpectrum being updated properly??

#-------------------------------------------------------------------------------
class AverageNtermModel(Model):
  """Average Nterm model"""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "AVG-N" 
    self.ntermProbDict = {}
    self.precursorMzModel = precursorMzModel

  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    ntermProbDict, ctermProbDict = calcAverageProbs(psmMatchCollection)
    self.ntermProbDict = ntermProbDict
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      peak.i = self.ntermProbDict[peak.cterm]

#-------------------------------------------------------------------------------
class AverageNtermAdjusterModel(AdjusterModel):
  """Used to test the Average Nterm model with AdjusterModel"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self, precursorMzModel)
    self.shortname = "AVG-N-ADJ" 
    self.ntermProbDict = {}

  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    ntermProbDict, ctermProbDict = calcAverageProbs(psmMatchCollection)
    self.ntermProbDict = ntermProbDict
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      peak.i = self.ntermProbDict[peak.cterm]
    AdjusterModel.adjustIntensityInfo(self, spectrum)

#-------------------------------------------------------------------------------
class AverageCtermModel(Model):
  """Average Cterm model"""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "AVG-C" 
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    ntermProbDict, ctermProbDict = calcAverageProbs(psmMatchCollection)
    self.ctermProbDict = ctermProbDict
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      peak.i = self.ctermProbDict[peak.nterm]

#-------------------------------------------------------------------------------
class AverageCtermAdjusterModel(AdjusterModel):
  """Average Cterm model"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self)
    self.shortname = "AVG-C-ADJ" 
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    ntermProbDict, ctermProbDict = calcAverageProbs(psmMatchCollection)
    self.ctermProbDict = ctermProbDict
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      peak.i = self.ctermProbDict[peak.nterm]
    AdjusterModel.adjustIntensityInfo(self, spectrum)


#-------------------------------------------------------------------------------
class AverageModel(Model):
  """Average model"""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "AVG" 
    self.ntermProbDict = {}
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    ntermProbDict, ctermProbDict = calcAverageProbs(psmMatchCollection)
    self.ntermProbDict = ntermProbDict
    self.ctermProbDict = ctermProbDict
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      peak.i = self.ntermProbDict[peak.cterm] * self.ctermProbDict[peak.nterm] 


#-------------------------------------------------------------------------------
class AverageAdjusterModel(AdjusterModel):
  """Average model"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self)
    self.shortname = "AVG-ADJ" 
    self.ntermProbDict = {}
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    ntermProbDict, ctermProbDict = calcAverageProbs(psmMatchCollection)
    self.ntermProbDict = ntermProbDict
    self.ctermProbDict = ctermProbDict
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getBYpeaks():
      peak.i = self.ntermProbDict[peak.cterm] * self.ctermProbDict[peak.nterm] 
    AdjusterModel.adjustIntensityInfo(self, spectrum)

#-------------------------------------------------------------------------------
# do a linear regression (via matlab) on the two-d aa statMatrix and produces
# a dict where the keys are aas and the values are probabilities of cleavage
def calcPairProbs(statsMatrix, probDict):

  conditionProbFileName = "input.dat"
  probFileName = "matlab.output" # FIX this is a hack. the matlab routine
  # has this hardcoded into it. must be a way around this.

  print_counts(statsMatrix, conditionProbFileName, "prob", labels=False)
  matlabCmd = """ matlab -r "probs_from_cond_prob('%s'); quit;" """ % \
       conditionProbFileName
  retcode = os.system(matlabCmd)
  print "Return code = %i" % retcode

  # FIX this assumes that the fields are in alphabetical order by aa
  probFile = open(probFileName, 'r')
  idx = 0
  for line in probFile: 
    probDict[aas[idx]] = float(line.rstrip('\n')) 
    idx += 1
    if idx > len(aas): 
      raise SystemExit, "Too many aas in matlab output %s" % probFile 
  probFile.close()
  
#-------------------------------------------------------------------------------
class PairNtermModel(Model):
  """N-term pair model"""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "PAIR-N"
    self.ntermProbDict = {}
    self.precursorMzModel = precursorMzModel

  # method for training the model on a collection of psms
  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    ntermStats, ctermStats = calcStats(psmMatchCollection, "prob")
    calcPairProbs(ntermStats, self.ntermProbDict)
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    for peak in richSpec.getPeaks():
      # prob dict is probability of cleaving nterm or cterm to residue
      # thus we examine the residue in cterm of the cleavage site.
      peak.i = self.ntermProbDict[peak.cterm]

#-------------------------------------------------------------------------------
class PairNtermAdjusterModel(AdjusterModel):
  """Used to test the Pair Nterm model with AdjusterModel"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self, precursorMzModel)
    self.shortname = "PAIR-N-ADJ" 
    self.ntermProbDict = {}

  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    ntermStats, ctermStats = calcStats(psmMatchCollection, "prob")
    calcPairProbs(ntermStats, self.ntermProbDict)
  
  def addIntensityInfo(self, spectrum):
    for peak in spectrum.getPeaks():
      # prob dict is probability of cleaving nterm or cterm to residue
      # thus we examine the residue in cterm of the cleavage site.
      peak.i = self.ntermProbDict[peak.cterm]
    AdjusterModel.adjustIntensityInfo(self, spectrum)

#-------------------------------------------------------------------------------
class PairCtermModel(Model):
  """C-term pair model"""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "PAIR-C"
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  # method for training the model on a collection of psms
  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    ntermsStats, ctermStats = calcStats(psmMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDict)
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    for peak in richSpec.getPeaks():
      # as PairNtermModel, but the opposite
      peak.i = self.ctermProbDict[peak.nterm]

#-------------------------------------------------------------------------------
class PairCtermAdjusterModel(AdjusterModel):
  """C-term pair model"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self)
    self.shortname = "PAIR-C-ADJ"
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  # method for training the model on a collection of psms
  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    ntermsStats, ctermStats = calcStats(psmMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDict)
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    for peak in richSpec.getPeaks():
      # as PairNtermModel, but the opposite
      peak.i = self.ctermProbDict[peak.nterm]
    AdjusterModel.adjustIntensityInfo(self, richSpec)

#-------------------------------------------------------------------------------
class PairModel(Model):
  """N-term and C-term pair model"""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "PAIR"
    self.ntermProbDict = {}
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  # trains the model on a collection of psms
  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    ntermStats, ctermStats = calcStats(psmMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDict)
    calcPairProbs(ntermStats, self.ntermProbDict)
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    for peak in richSpec.getPeaks():
      # calculate probability of cleaving C-term to N-term residue
      # and N-term to C-term residue, assuming independence
      peak.i = self.ntermProbDict[peak.cterm] * self.ctermProbDict[peak.nterm]
      
#-------------------------------------------------------------------------------
class PairAdjusterModel(AdjusterModel):
  """N-term and C-term pair model"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self)
    self.shortname = "PAIR-ADJ"
    self.ntermProbDict = {}
    self.ctermProbDict = {}
    self.precursorMzModel = precursorMzModel

  # trains the model on a collection of psms
  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    ntermStats, ctermStats = calcStats(psmMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDict)
    calcPairProbs(ntermStats, self.ntermProbDict)
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    for peak in richSpec.getBYpeaks():
      # calculate probability of cleaving C-term to N-term residue
      # and N-term to C-term residue, assuming independence
      try:
        peak.i = self.ntermProbDict[peak.cterm] * self.ctermProbDict[peak.nterm]
      except:
        print >>sys.stderr, "Could not find cterm %c or nterm %c in probdict" %\
          (peak.cterm, peak.nterm)
    AdjusterModel.adjustIntensityInfo(self, richSpec)

#-------------------------------------------------------------------------------
BY_INTENSITY            = 50.0
BY_FLANK_INTENSITY      = 25.0
A_INTENSITY             = 10.0
NEUTRAL_LOSS_INTENSITY  = 10.0
FLANK_MZ_OFFSET         = 1.0
#-------------------------------------------------------------------------------
class SequestModel(Model):
  """Sequest model"""

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.shortname = "SEQUEST" 

  def train(self, psmMatchCollection):
    pass
  
  def addIntensityInfo(self, spectrum):
    flankingPeaks = []
    for peak in spectrum.getPeaks():
      if peak.isBpeak() or peak.isYpeak():       
        peak.i = BY_INTENSITY
        fl = RichPeak(mz=(peak.mz - FLANK_MZ_OFFSET), i=BY_FLANK_INTENSITY)
        fr = RichPeak(mz=(peak.mz + FLANK_MZ_OFFSET), i=BY_FLANK_INTENSITY)
        fl.type = fr.type = "f"
        flankingPeaks.append(fl)
        flankingPeaks.append(fr)
      elif peak.isApeak():                       
        peak.i = A_INTENSITY
      elif peak.isNH3peak() or peak.isH2Opeak(): 
        peak.i = NEUTRAL_LOSS_INTENSITY

    for peak in flankingPeaks: spectrum.addPeak(peak)

#START
#-------------------------------------------------------------------------------
class PairAdjusterMprotonModel(AdjusterModel):
  """N-term and C-term pair model, with mobile proton information"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self)
    self.shortname = "PAIR-MP"

    self.ntermProbDicts = {}
    self.ctermProbDicts = {}

    for key in ("Mobile", "NonMobile", "PartiallyMobile"):
      self.ntermProbDicts[key] = {}
      self.ctermProbDicts[key] = {}

    self.precursorMzModel = precursorMzModel


  # trains the model on a collection of psms
  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    mobility = "Non-mobile"
    mobileMatchCollection = PsmMatchCollection()
    partiallyMobileMatchCollection = PsmMatchCollection()
    nonMobileMatchCollection = PsmMatchCollection()
    for match in psmMatchCollection.getMatches():
      Rs = Mass.countR(match.peptide)
      Ks = Mass.countK(match.peptide)
      if match.spectrum["charge"] > Rs + Ks:
        mobileMatchCollection.appendMatch(match)
      elif match.spectrum["charge"] > Rs and match.spectrum["charge"] <= Rs + Ks:
        partiallyMobileMatchCollection.appendMatch(match)
      elif match.spectrum["charge"] <= Rs:
        nonMobileMatchCollection.appendMatch(match)
      else: 
        raise SystemExit, "Weirdness!"

    ntermStats, ctermStats = calcStats(mobileMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDicts["Mobile"])
    calcPairProbs(ntermStats, self.ntermProbDicts["Mobile"])

    ntermStats, ctermStats = calcStats(nonMobileMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDicts["NonMobile"])
    calcPairProbs(ntermStats, self.ntermProbDicts["NonMobile"])
    
    ntermStats, ctermStats = calcStats(partiallyMobileMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDicts["PartiallyMobile"])
    calcPairProbs(ntermStats, self.ntermProbDicts["PartiallyMobile"])

    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    Rs = Mass.countR(richSpec.peptide)
    Ks = Mass.countK(richSpec.peptide)
    if richSpec["charge"] > Rs + Ks: key = "Mobile" 
    elif richSpec["charge"] > Rs and richSpec["charge"] <= Rs + Ks: key = "PartiallyMobile"
    elif richSpec["charge"] <= Rs: key = "NonMobile"
    else: raise SystemExit, "Weirdness!"

    for peak in richSpec.getPeaks():
      # calculate probability of cleaving C-term to N-term residue
      # and N-term to C-term residue, assuming independence
      peak.i = self.ntermProbDicts[key][peak.cterm] * self.ctermProbDicts[key][peak.nterm]

    AdjusterModel.adjustIntensityInfo(self, richSpec)

#-------------------------------------------------------------------------------
class PairProductAdjusterModel(AdjusterModel):
  """N-term and C-term pair model. Calculates joint probability of peak
  given all other peaks"""

  def __init__(self, precursorMzModel=None):
    AdjusterModel.__init__(self)
    self.shortname = "PAIR-PR-ADJ"
    self.ntermStats = {}
    self.ctermStats = {}
    self.precursorMzModel = precursorMzModel

  # trains the model on a collection of psms
  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    self.ntermStats, self.ctermStats = calcStats(psmMatchCollection, "prob")
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    peaks = richSpec.getPeaks()
    for idx in range(len(peaks)):
      peak = peaks[idx]
      peakHeight = 1.0
      for otherIdx in range(len(peaks)):
        if idx == otherIdx: continue
        otherPeak = peaks[otherIdx]
        # FIX
        ntermProb = self.ntermStats[peak.cterm][otherPeak.cterm]['AVG_PROB']
        ctermProb = self.ctermStats[peak.nterm][otherPeak.nterm]['AVG_PROB']
        peakHeight *= ntermProb   
        peakHeight *= ctermProb   
      peak.i = peakHeight
    AdjusterModel.adjustIntensityInfo(self, richSpec)


#-------------------------------------------------------------------------------
class PairChargeModel(PairModel):

  def __init__(self, precursorMzModel=None):
    Model.__init__(self)
    self.data = {}
    self.type = "Pair model with charge information"
    self.shortname = "PAIR-CHR"
    self.ntermProbDict = {}
    self.precursorMzModel = precursorMzModel

  # method for training the model on a collection of psms
  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)

    # Use some independently generated data if no psmMatchCollection
    if psmMatchCollection == None:
      ntermProbDict =  [ 0.0464, 0.0537, 0.0519, 0.0474, 0.0517, 0.0606, \
        0.0551, 0.0505, 0.0211, 0.0466, 0.0492, 0.0526, 0.0981, 0.0470, \
        0.0154, 0.0501, 0.0512, 0.0491, 0.0507, 0.0478 ] 
      for idx in range(len(aas)): self.ntermProbDict[aas[idx]]=ntermProbDict[idx]
      return

    # Really train the model if we have PSMs
    nterms_stats, cterm_stats = calcStats(psmMatchCollection, "prob")
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    for peak in richSpec.getPeaks():
      # prob dict is probability of cleaving nterm or cterm to residue
      # thus we examine the residue in cterm of the cleavage site.
      peak.i = self.ntermProbDict[peak.cterm]

# START
#-------------------------------------------------------------------------------
# class PairMobileProtonModel(PairModel):
#    self.type = "A group of pair models each trained on peptides with different proton mobility"

#-------------------------------------------------------------------------------
B_RATIO = 0.5
NH3_LOSS_RATIO = 0.5
H2O_LOSS_RATIO = 0.5
CO_LOSS_RATIO = 0.5
ISOTOPE_RATIO = 0.1
PLUS_TWO_RATIO = 0.2


#-------------------------------------------------------------------------------
class AsmsModel(AdjusterModel):
  """Model solely for ASMS"""

  def __init__(self):
    AdjusterModel.__init__(self)
    self.shortname = "ASMS"
    self.ntermProbDict = {}
    self.ctermProbDict = {}
    self.ntermStats = {}
    self.ctermStats = {}
    self.chargeAllocProbs = {}

  # trains the model on a collection of psms
  def train(self, psmMatchCollection):
    AdjusterModel.train(self, psmMatchCollection)
    ntermStats, ctermStats = calcStats(psmMatchCollection, "prob")
    calcPairProbs(ctermStats, self.ctermProbDict)
    calcPairProbs(ntermStats, self.ntermProbDict)

    self.ntermStats = ntermStats
    self.ctermStats = ctermStats

    # train b-y ion charge allocation
    for match in psmMatchCollection.getMatches(): 
      richspec = RichSpectrum()
      richspec.addPeptideIons(match.peptide)
      richspec.addIntensitiesFromSpectrum(match.spectrum)
      for peak in richspec.getPeaks():
        if not peak.isBpeak(): continue
        if peak.i == 0 and peak.partnerPeak.i == 0: continue
        prob = peak.i / (peak.i + peak.partnerPeak.i) 
        key = (Mass.countBasic(peak.nterm), Mass.countBasic(peak.cterm), \
			Mass.countBasic(peak.frag), Mass.countBasic(peak.partnerPeak.frag))
        if self.chargeAllocProbs.has_key(key): 
          self.chargeAllocProbs[key].append(prob)
        else: self.chargeAllocProbs[key] = [ prob ]

    for key in self.chargeAllocProbs.keys(): 
      self.chargeAllocProbs[key] = \
        Numeric.average( self.chargeAllocProbs[key] )
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
   
    for peak in richSpec.getBYpeaks():
      # calculate probability of cleaving C-term to N-term residue
      # and N-term to C-term residue, assuming independence
      try:
        peak.i = self.ntermProbDict[peak.cterm] * self.ctermProbDict[peak.nterm]
      except:
        print >>sys.stderr, "Could not find cterm %c or nterm %c in probdict" %\
          (peak.cterm, peak.nterm)

    # account for b-y charge allocation
    for peak in richSpec.getPeaks():
      if not peak.isBpeak(): continue
      peakKey = (Mass.countBasic(peak.nterm), Mass.countBasic(peak.cterm), \
		      Mass.countBasic(peak.frag), Mass.countBasic(peak.partnerPeak.frag))
      byRatio = 1 # default is 1:1 b:y ratio
      if self.chargeAllocProbs.has_key(peakKey): 
        byRatio = self.chargeAllocProbs[peakKey]
      peak.i *= byRatio 

    #----
    # for peak in richSpec.getBYpeaks(): 
    #  if peak.isBpeak():
    #	peak.i = peak.i * B_RATIO
    #----
    for peak in richSpec.getNH3peaks(): 
      peak.i = peak.parentPeak.i * NH3_LOSS_RATIO
    for peak in richSpec.getApeaks():
      peak.i = peak.parentPeak.i * CO_LOSS_RATIO
    for peak in richSpec.getH2Opeaks():
      peak.i = peak.parentPeak.i * H2O_LOSS_RATIO

    # add isotope distribution
    for peak in richSpec.getPeaks():
      if peak.isIsotopePeak(): peak.i = peak.parentPeak.i * ISOTOPE_RATIO

    # add +2 charge info
    for peak in richSpec.getPeaks():
      if peak.z == 2: peak.i = peak.parentPeak.i * PLUS_TWO_RATIO

    AdjusterModel.adjustIntensityInfo(self, richSpec)
    # for peak in richSpec.getPeaks(): peak.i = peak.i * 100000

#-------------------------------------------------------------------------------
class GmtkF(Model):
  """Model that takes a model-F params file as input"""

  def __init__(self):
    self.shortname = "GMTK-F-SQRT-BP"

    self.binValues = []

    self.bNtermProbDict = {}
    self.bCtermProbDict = {}
    self.bNalpha = 0.0

    self.yNtermProbDict = {}
    self.yCtermProbDict = {}
    self.yNalpha = 0.0

  def parseCPT(self, lines, idx):
    length, width = map(int, lines[idx+2].split(' ')[0:2])
    cpt = []
    cptStartOffset = 4 
    for cptIdx in range(idx + cptStartOffset, idx + cptStartOffset + length):
      row = lines[cptIdx].rstrip('\n').split(' ')
      row = map(float, row[0:len(row)-1])
      cpt.append(row)
    return cpt

  def parseMixture(self, lines, idx): # TODO need to change for more than 
    # two mixtures
    myIdx = idx + 3
    row = lines[myIdx].rstrip('\n').split(' ')
    return float(row[0])

  def aa2probDict(self, cpt): 
    dict = {}
    aas = Mass.CanonicalAminos
    for idx in range(len(cpt)):
      row = cpt[idx]
      dict[ aas[idx] ] =  Numeric.average( \
          [ self.binValues[i] * row[i] for i in range(len(row)) ]) 
    return dict

  # trains the model on a collection of psms
  def train(self, gmtkParms):
    file = open(gmtkParms)
    lines = file.readlines()
    file.close()
    for idx in range(len(lines)):
      line = lines[idx]
      if line.startswith("iBintRR_nA"): 
        cpt = self.parseCPT(lines, idx)
        self.bNtermProbDict = self.aa2probDict(cpt)
      elif line.startswith("iYintRR_nA"): 
        cpt = self.parseCPT(lines, idx)
        self.yNtermProbDict = self.aa2probDict(cpt)
      elif line.startswith("iBintRR_cA"): 
        cpt = self.parseCPT(lines, idx)
        self.bCtermProbDict = self.aa2probDict(cpt)
      elif line.startswith("iYintRR_cA"): 
        cpt = self.parseCPT(lines, idx)
        self.yCtermProbDict = self.aa2probDict(cpt)
      elif line.startswith("iBmix_CPT"): 
        self.bNalpha = self.parseMixture(lines, idx)
      elif line.startswith("iYmix_CPT"): 
        self.yNalpha = self.parseMixture(lines, idx)
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
   
    for peak in richSpec.getBYpeaks():
      # calculate probability of cleaving C-term to N-term residue
      # and N-term to C-term residue, assuming independence
      if peak.type == 'b':
        if peak.nterm == '_' : continue
        peak.i = float(self.bNtermProbDict[peak.nterm])
        peak.i *= float(self.bNalpha)
        peak.i += float(self.bCtermProbDict[peak.cterm]) * (1.0-float(self.bNalpha))
      elif peak.type == 'y':
        peak.i = float(self.yNtermProbDict[peak.nterm])
        peak.i *= float(self.yNalpha)
        peak.i += float(self.yCtermProbDict[peak.cterm]) * (1.0-float(self.yNalpha))
      #------------
      # try:
      #   peak.i = self.ntermProbDict[peak.cterm] * self.ctermProbDict[peak.nterm]
      # except:
      #   print >>sys.stderr, "Could not find cterm %c or nterm %c in probdict" %\
      #   (peak.cterm, peak.nterm)
      # account for b-y charge allocation
      #------------


#----
# This looks at the top peak, and compares all peaks to that peak. Didn't work that 
# well for ASMS
#----
# find what should be the highest peak
#    topCi = 0.0
#    topCpeak = None
#    topNi = 0.0
#    topNpeak = None
#
#
#    for peak in richSpec.getBYpeaks():
#      # calculate probability of cleaving C-term to N-term residue
#      # and N-term to C-term residue, assuming independence
#      try:
#	if self.ntermProbDict[peak.cterm] > topNi:
#	  topNpeak = peak
#	  topNi = self.ntermProbDict[peak.cterm]
#	if self.ctermProbDict[peak.nterm] > topCi: 
#	  topCpeak = peak
#	  topCi = self.ctermProbDict[peak.nterm]
#      except:
#        print >>sys.stderr, "Could not find cterm %c or nterm %c in probdict" %\
#          (peak.cterm, peak.nterm)
#
#    # set all peaks relative to that highest peak 
#    for peak in richSpec.getBYpeaks():
#      try:
#	# FIX should not access AVG_PROB directly
#	# FIX should we factor in the height of the topPeak at all?
#        peak.i = self.ntermStats[peak.cterm][topNpeak.cterm]['AVG_PROB'] * \
#	    self.ctermStats[peak.nterm][topCpeak.nterm]['AVG_PROB']
#      except: pass
#----

class AverageChargeAllocationModel(AverageModel):

  def __init__(self, precursorMzModel=None):
    AverageModel.__init__(self, precursorMzModel)
    self.data = {}
    self.type = "Average model with charge allocation information"
    self.shortname = "PAIR-BY"
    self.probs = {}

  # method for training the model on a collection of psms
  def train(self, psmMatchCollection):
    AverageModel.train(self, psmMatchCollection)
    for match in psmMatchCollection.getMatches(): 
      richspec = RichSpectrum()
      richspec.addPeptideIons(match.peptide)
      richspec.addIntensitiesFromSpectrum(match.spectrum)
      for peak in richspec.getPeaks():
        if not peak.isBpeak(): continue
        if peak.i == 0 and peak.partnerPeak.i == 0: continue
        prob = peak.i / (peak.i + peak.partnerPeak.i) 
        key = (Mass.countBasic(peak.frag), Mass.countBasic(peak.partnerPeak.frag))
        if self.probs.has_key(key): self.probs[key].append(prob)
        else: self.probs[key] = [ prob ]

    for key in self.probs.keys(): self.probs[key] = Numeric.average( self.probs[key] )

  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    AverageModel.addIntensityInfo(self, richSpec)
    for peak in richSpec.getPeaks():
      if not peak.isBpeak(): continue
      peakKey = (Mass.countBasic(peak.frag), Mass.countBasic(peak.partnerPeak.frag))
      byRatio = 1 # default is 1:1 b:y ratio
      if self.probs.has_key(peakKey): byRatio = self.probs[peakKey]
      peak.i *= byRatio 



#-------------------------------------------------------------------------------
class PairChargeAllocationModel(PairAdjusterModel):

  def __init__(self, precursorMzModel=None):
    PairAdjusterModel.__init__(self, precursorMzModel)
    self.data = {}
    self.type = "Pair model with charge allocation information"
    self.shortname = "PAIR-BY"
    self.probs = {}

  # method for training the model on a collection of psms
  def train(self, psmMatchCollection):
    PairAdjusterModel.train(self, psmMatchCollection)
    for match in psmMatchCollection.getMatches(): 
      richspec = RichSpectrum()
      richspec.addPeptideIons(match.peptide)
      richspec.addIntensitiesFromSpectrum(match.spectrum)
      for peak in richspec.getPeaks():
        if not peak.isBpeak(): continue
        if peak.i == 0 and peak.partnerPeak.i == 0: continue
        prob = peak.i / (peak.i + peak.partnerPeak.i) 
        key = (Mass.countBasic(peak.frag), Mass.countBasic(peak.partnerPeak.frag))
        if self.probs.has_key(key): self.probs[key].append(prob)
        else: self.probs[key] = [ prob ]

    for key in self.probs.keys(): 
      self.probs[key] = Numeric.average( self.probs[key] )

  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    PairAdjusterModel.addIntensityInfo(self, richSpec)
    for peak in richSpec.getPeaks():
      if not peak.isBpeak(): continue
      peakKey = (Mass.countBasic(peak.frag), Mass.countBasic(peak.partnerPeak.frag))
      byRatio = 1 # default is 1:1 b:y ratio
      if self.probs.has_key(peakKey): byRatio = self.probs[peakKey]
      peak.i *= byRatio 

#-------------------------------------------------------------------------------
class PairNeutralLossModel(PairModel):

  def __init__(self, precursorMzModel=None):
    PairModel.__init__(self)
    self.data = {}
    self.type = "Pair model with charge, charge allocation and neutral losses."
    self.shortname = "PAIR-NEU"
    self.ntermProbDict = {}
    self.precursorMzModel = precursorMzModel

  # method for training the model on a collection of psms
  def train(self, psmMatchCollection):
    Model.train(self, psmMatchCollection)
    # Use some independently generated data if no psmMatchCollection
    if psmMatchCollection == None:
      ntermProbDict =  [ 0.0464, 0.0537, 0.0519, 0.0474, 0.0517, 0.0606, \
        0.0551, 0.0505, 0.0211, 0.0466, 0.0492, 0.0526, 0.0981, 0.0470, \
        0.0154, 0.0501, 0.0512, 0.0491, 0.0507, 0.0478 ] 
      for idx in range(len(aas)): self.ntermProbDict[aas[idx]]=ntermProbDict[idx]
      return

    # Really train the model if we have PSMs
    nterms_stats, cterm_stats = calcStats(psmMatchCollection, "prob")
    
  # add the intensity info to the RichSpectrum richSpec
  def addIntensityInfo(self, richSpec):
    for peak in richSpec.getPeaks():
      # prob dict is probability of cleaving nterm or cterm to residue
      # thus we examine the residue in cterm of the cleavage site.
      peak.i = self.ntermProbDict[peak.cterm]
    richSpec.sumNormalizeIntensity() # TODO put somewhere else?
 

#-------------------------------------------------------------------------------
class StrawModel(Model):

  #----
  def __init__(self, gmtkModel=None, paramPrefix=None):
    self.shortname = "FRAG-STRAW"


  #---------------------------------------------------------------------------
  # add the intensity info to the RichSpectrum richSpec
  #---------------------------------------------------------------------------
  def addIntensityInfo(self, richSpec, type='mode'):
    yPeaks = richSpec.getPeaksOfType(Y_KEY)
    startingIdx = int(len(yPeaks)/2.0) + 1

    diffs = []
    for idx in range(1,int(len(yPeaks)/2.0)):
      diffs.append(-idx)
      diffs.append(idx)

    indices = [startingIdx]
    for diff in diffs:
      indices.append(startingIdx + diff)
      
    indices.reverse()
    yPeaks.sort(lambda x,y: cmp(x.mz, y.mz))
    # set the peaks to the appropriate intensity
    for index_index in range(len(indices)):
      # print "%i - %i / %i" % (index_index, indices[index_index], len(yPeaks)+1)
      yPeaks[indices[index_index]-1].i = index_index


#-------------------------------------------------------------------------------
class GmtkModel(Model):

  def __init__(self, gmtkModel=None, paramPrefix=None):
    self.shortname = "GMTK"
    self.gmtkModel = gmtkModel
    self.paramPrefix = paramPrefix
    self.gmtkProbs = None
    self.mzInt2gmtkRegion = [ 0, 1, 1, 2, 2, 3, 4, 4, 1, 1, 0, 99 ]

  #---------------------------------------------------------------------------
  def peakRegion(self, spectrum, peak):
    pmz = HmmData.mz2int(peak.mz, spectrum["precursor mz"])
    return self.mzInt2gmtkRegion[pmz]

  #---------------------------------------------------------------------------
  # method for training the model on a collection of psms
  #---------------------------------------------------------------------------
  def train(self):
    self.gmtkProbs = GMTKprobs.GMTKprobCollection(self.gmtkModel)
    self.gmtkProbs.parse(self.paramPrefix)
    
  #---------------------------------------------------------------------------
  # add the intensity info to the RichSpectrum richSpec
  #---------------------------------------------------------------------------
  def addIntensityInfo(self, richSpec, type='mode'):
    for peak in richSpec.getPeaks():
      region = self.peakRegion(richSpec, peak)
      peak.i = self.gmtkProbs.calcProb(\
          peak.nterm, peak.cterm, region, peak.type, type=type)
    #richSpec.sumNormalizeIntensity() # TODO put somewhere else?
 
#-------------------------------------------------------------------------------
# Main
#-------------------------------------------------------------------------------
if __name__ == '__main__':
  gmtkModel = "singleIon-a"
  paramPrefix = "large"
  m = GmtkModel(gmtkModel, paramPrefix)
  m.train()
  fh = open("GMTK-large.model", "w")
  cPickle.dump(m, fh)
  fh.close()
  #print >>sys.stderr, "Loading cPickledgmtk"
  #m = cPickle.load(open("cPickledgmtk"))
  #m.mzInt2gmtkRegion = [ 0, 1, 1, 2, 2, 3, 4, 4, 1, 1, 0, 99 ]

  #r = RichSpectrum()
  #r.addPeptideIons(sys.argv[1], neutrals=True, multiCharge=True, isotopes=False)
  #r.addPeptideIons(sys.argv[1]) #, neutrals=True, multiCharge=True, isotopes=False)
  #print >>sys.stderr, "Predicting intensities"
  #m.addIntensityInfo(r, type='mode')
  #print r






