#-------------------------------------------------------------------------------
# Copyright 2005 University of Washington. All rights reserved.
#-------------------------------------------------------------------------------
# $Author: aklammer $ 
#-----------------
# $Revision: 1.5 $
#-----------------
import math,sys,os
G_platt_fname = "plattsvm"
NUM_FEATURES = 500

def dot_product(vector1, vector2):
  if len(vector1) != len(vector2):
    print >>sys.stderr, "Vectors in dot_product of unequal length %i %i!" % \
        (len(vector1), len(vector2))
    sys.exit(-1)
  
  dot_product = 0.0
  for idx in range(len(vector1)):
    dot_product += vector1[idx] * vector2[idx]

  return dot_product

def l2_norm_square(vector1, vector2):
  if len(vector1) != len(vector2):
    print >>sys.stderr, "Vectors in l2_norm_square of unequal length! %i %i" % \
        (len(vector1), len(vector2))
    sys.exit(-1)

  # print str(vector1)
  # print str(vector2)
  dot_11 = dot_product(vector1, vector1)
  dot_12 = dot_product(vector1, vector2)
  dot_22 = dot_product(vector2, vector2)

  return dot_11 - 2*dot_12 + dot_22

class LinearSvm:
  """Container object for information regarding a linear Svm.
  """

  def __init__(self):
    self.alphas   = []
    self.labels   = []
    self.svectors = []
    self.feature_names = []
    self.weights = []
    self.bias = 0.0

  def classify(self, vector):
    disc = 0.0
    for idx in range(len(self.svectors)):
      disc += self.alphas[idx] * dot_product(self.svectors[idx], vector)
    disc += self.bias
    return disc

  def calc_discriminant(self, vector):
    return self.classify(vector)
  
  def parseAlphas(self, line):
    self.alphas = map(float, line.split('=')[1].split(' '))

  def parseVector(self, line):
    fields = line.split()
    label, dependent = fields[0].split(',')
    vector = [0] * NUM_FEATURES
    for field in fields[1:]:
      idx, value = field.split(':')
      vector[int(idx)] = float(value)
    self.labels.append(label)
    self.svectors.append(vector)

  def parseFromFile(self, filename):
    fh = open(filename)
    for line in fh:
      line = line.rstrip('\n')
      if line.startswith('#b'):
        self.bias = float(line.split('=')[1])
      if line.startswith('#alpha'):
        self.parseAlphas(line)
      if line.startswith('#'): continue
      self.parseVector(line)
    fh.close()
  
  def populateWeights(self):
    weights = [0.0] * len(self.svectors[0])
    for vector_idx in range(len(self.svectors)):
      for feature_idx in range(len(self.svectors[vector_idx])):
        weights[feature_idx] += self.alphas[vector_idx] * \
          self.svectors[vector_idx][feature_idx]
    self.weights = weights  

class Svm:
  """Container object for information regarding a Platt trained Gaussian Svm.
    K(x,y) = exp( - gamma * ||x - y||**2)  
  """


  def __init__(self):
    self.A = 0.0
    self.B = 0.0
    self.translation = []
    self.scale = []
    self.gamma = 0.0
    self.alphas   = []
    self.labels   = []
    self.svectors = []
    self.feature_names = []

  def calc_discriminant(self, vector):
    """ Calculates the discriminant for the vector <vector>"""
    if len(vector) != len(self.svectors[0]):
      print >>sys.stderr, "Vectors in calc_discriminant of unequal length %i %i!" % \
          (len(vector), len(self.svectors[0]))
      sys.exit(-1)

    for idx in range(len(vector)):
      vector[idx] -= self.translation[idx]
      vector[idx] /= self.scale[idx]

    discriminant = 0.0
    for svector_idx in range(len(self.svectors)):
      svector = self.svectors[svector_idx]
      exponential = math.exp(- self.gamma * l2_norm_square(vector, svector))
      alpha = self.alphas[svector_idx]
      discriminant += exponential * alpha # * self.labels[svector_idx]
  
    return discriminant + self.bias

  def calc_probability(self, vector):
    """ Calculates the probability from the sigmoid fitted as by Platt
    for discriminant <discriminant> using 1/(1+exp(A*f(x)+B)"""

    discriminant  = self.calc_discriminant(vector)
    probability = 1 / (1 + math.exp(self.A * discriminant  + self.B))
    return probability

  def write(self, rescale=None, classifier=None, output="svm.svmobject"):

    self.translation = rescale.translation
    self.scale  = rescale.scale.tolist()

    if 'Gaussian' not in classifier.model.svdata.kernel.dump():
      print >>sys.stderr, "Right now we only work with Gaussian kernels!"
      sys.exit(-1)

    self.gamma  = classifier.model.svdata.kernel.gamma
    self.bias   = classifier.model.b
    self.alphas = classifier.model.alpha
    classifier.model.svdata.save('svs', format='csv')
    svector_file = open('svs', 'r')
    svectors = []
    labels  = []
    for line in svector_file:
      if line.startswith('#'): continue
      fields = line.split(',')

      label = int(fields[1])
      labels.append(label)

      vector = map(float, fields[2:])
      svectors.append(vector)

    self.svectors = svectors
    self.labels = labels

    print >>sys.stderr, "Outputting SVM\t", time.asctime()
    file = open(output, 'w')
    cPickle.dump(self, file)
    file.close()
    
    print >>sys.stderr, "Finished processing\t", time.asctime()



def get_platt_svm_fname():
  path = os.path.dirname(__file__) + os.path.sep + G_platt_fname
  return path
