#---------------------------------------------------------------------------
# Name: generatePeptides.py
#---------------------------------------------------------------------------
# Author: $Author: aklammer $
#---------------------------------------------------------------------------
# Revision: $Revision: 1.52 $
#---------------------------------------------------------------------------
# Inputs:
#  - a .fasta file from which to generate peptides
#  - a <collection-prefix>.scoll|pickledpsm object with spectra 
#    for which to generate candidate peptides 
#--------------------------------------------------------------------------
# Outputs:
#  - a directory <fasta-prefix>,<collection-prefix>.pepdir 
#    containing a files for each spectrum with its peptides:
#      - spectrum-id.cand - candidate peptides for matches
#  - an output file <fasta-prefix>,<collection-prefix>.peps 
#     that maps the spectra to their peptide files: 
#     spectrum id \t candidate peptide file \n
#---------------------------------------------------------------------------
import os, sys

#---------------------------------------------------------------------------
if __name__ == "__main__":
  if len(sys.argv) < 2:
    print >>sys.stderr, """
Usage: %s <fasta-prefix> <collection-prefix> 
    <fasta-file> - A .fasta file for generating null and candidate peptides
    <collection-file> 
             - The pickled SpectrumCollection (with suffix .pickledscoll) or 
               PsmMatchCollection (with suffix .pickledpsm) object that 
               will be searched.
""" % os.path.basename(sys.argv[0])
    sys.exit(1)

#---------------------------------------------------------------------------
from hmmParams import *
try: import Bio.Fasta
except: print >>sys.stderr, "Could not import Bio.Fasta"
import random
import Mass
import SpectrumCollection, Spectrum, MatchCollection, Match
import cPickle
import Crux

#---------------------------------------------------------------------------
def candPeptideDict(seqs):
  """
  return a dict mapping mass:[peptides] 
  """

  # find tryptic seqs
  z = 2
  mass2seq = {}
  for seq in seqs:
    for start in range(len(seq)):
      mass = Mass.mono_mass('H2O') + Mass.mono_mass('H_plus') * z
      for end in range(start, len(seq)):
        mass += Mass.mono_mass(seq[end])
        massInt = int(mass)
        if (massInt < MIN_MASS): continue
        if (MAX_MASS < massInt): break
        if mass2seq.has_key(mass):
          mass2seq[massInt].append(seq[start:end+1])
        else:
          mass2seq[massInt] = [ seq[start:end+1] ]

  return mass2seq


def candTrypticPeptideDict(fastaPrefix, massesAllowed):
  """
  return a dict mapping mass:[peptides] 
  uses only tryptic peptides
  """

  print >>sys.stderr, "Reading fasta file %s" % fastaPrefix
  fh = open(fastaPrefix + ".fasta")
  idx = 0
  trypticSeqs = []
  seq = ""

  # remove first line
  line = fh.readline()

  almostDone = False
  while(True):
    curLine = fh.readline() 
    if almostDone == True : break
    if curLine == "": 
      almostDone = True
    curLine = curLine.rstrip('\n')
    if not curLine.startswith('>') and not curLine == "":
      seq += curLine
      continue

    idx += 1
    if idx % 1000 == 0: print >>sys.stderr, "Reached sequence %i" % idx
    trypticSeq = []
    for i in range(len(seq)):
      trypticSeq.append(seq[i])
      if (seq[i] == 'K' or seq[i] == 'R'):
        if i < len(seq)-1:
          if seq[i+1] != 'P':
            if len(trypticSeq) >= MIN_PEPTIDE_LEN:
              trypticSequence = ''.join(trypticSeq)
              mass = Mass.peptide_mass(trypticSequence, z=2)
              massInt = int(mass + 0.5)
              if ( (massInt < MIN_MASS) or (MAX_MASS < massInt) or not \
                  massesAllowed.has_key(massInt)  or not \
                  Mass.canonicalOnly(trypticSeq)): 
                trypticSeq = []
                continue
              trypticSeqs.append(trypticSequence)
            trypticSeq = []

    # add any remaining seq
    if len(trypticSeq) >= MIN_PEPTIDE_LEN:
      trypticSequence = ''.join(trypticSeq)
      mass = Mass.peptide_mass(trypticSequence, z=2)
      massInt = int(mass + 0.5)
      if ( (massInt < MIN_MASS) or (MAX_MASS < massInt) or not \
          massesAllowed.has_key(massInt) or not \
          Mass.canonicalOnly(trypticSeq)): 
        seq = ""
        trypticSeq = []
        continue
      trypticSeqs.append(trypticSequence)
    seq = ""
  fh.close()
  # calc mass dict
  # FIX eliminate X's and other non-standard aas
  mass2seq = {}
  for peptide in trypticSeqs:
    massBin = int(Mass.peptide_mass(peptide, z=2) + 0.5)
    if mass2seq.has_key(massBin):
      if not mass2seq[massBin].has_key(peptide):
        mass2seq[massBin][peptide] = 1
    else:
      mass2seq[massBin] = { peptide : 1 }

  return mass2seq

#----------------------
def frequencyToRandDict(freqDict):
  """
  return a list of length PRECISION where values are the keys of input dict.
  Each key in input dict is present as a value in output dict proportional
  to its corresponding value in input freqDict.
  """

  idx = 0
  rand = []
  for key in freqDict:
    for subIdx in range(int(freqDict[key]*PRECISION)):
      rand.append(key)
      idx+=1
  if len(rand) < PRECISION: rand.append(rand[-1])
  return rand


def reverse(peptide):
  rev = ""
  for i in range(len(peptide)):
    rev+=peptide[len(peptide)-1-i]
  return rev

#--------------------------------------------------------------------
#--------------------------------------------------------------------
# generate a random tryptic 
# peptide with aa freq aaFreq within one of the masses in massesAllowed
#--------------------------------------------------------------------
def generateTrypticPeptide(randDict, kFreq=0.5, rFreq=0.5, massesAllowed=None, z=2):
  """
i aa:frequency 
i mass:exist
o peptide
  """
  kThresh = kFreq / (kFreq + rFreq)
   
  while(True):
    endingAA = ""
    if random.random() < kThresh: endingAA = 'K'
    else: endingAA = 'R'

    peptide = endingAA
    mass = Mass.peptide_mass(endingAA, z=2)

    for i in range(ITERATIONS):

      while mass < MAX_MASS:
        raa = randDict[ int(len(randDict)*random.random()) ] # Mass.randomAA()
        peptide += raa
        mass += Mass.mono_mass(raa)
        massInt = int(mass+0.5)
        if massesAllowed.has_key(massInt): 
          return reverse(peptide), massInt

      while MIN_MASS < mass and len(peptide) > 1:
        raa = peptide[-1]
        mass -= Mass.mono_mass(raa)
        peptide = peptide[0:len(peptide)-1] 

#------------------------------------------------------------------------
def nullPeptideDict(fastaFname, massLeft):
  """
  return a dict mapping mass:[peptides] 
  uses only tryptic peptides
  """

  #  # count aa frequencies
  #  aaCount = {}
  #  for aa in Mass.CanonicalAminos: aaCount[aa] = 0
  #  totalCount = 0
  #
  #  print >>sys.stderr, "Reading fasta file %s" % fastaFname
  #  fh = open(fastaFname)
  #  iterator = Bio.Fasta.Iterator(fh, Bio.Fasta.RecordParser())
  #  seq_idx = 0
  #  while(True):
  #    seq_idx += 1
  #    curRecord = iterator.next()
  #    if curRecord is None: break
  #    if seq_idx % 2 == 1: 
  #      print >>sys.stderr, "Reached sequence %i" % seq_idx
  #    else:
  #      continue
  #    seq = curRecord.sequence
  #    totalCount += len(seq)
  #    for aa_idx in range(len(seq)): 
  #      if aaCount.has_key(seq[aa_idx]): aaCount[seq[aa_idx]]+=1
  #
  #  fh.close()
  #
  # aaFrac = {} 
  # for key in aaCount.keys(): aaFrac[key] = float(aaCount[key])/totalCount
  # print aaCount

  #------------------------------
  # From the E. coli proteome
  #------------------------------
  aaFrac =  \
  {'A': 0.094323673260821958, 'C': 0.011797941455580699, 'E': 0.057979414555806991, \
   'D': 0.051430031389152496, 'G': 0.073840426308489665, 'F': 0.038983867435579238, \
   'I': 0.059763486385867579, 'H': 0.022862982699467113, 'K': 0.043702460033579096, \
   'M': 0.028060442368056063, 'L': 0.10642820643842617, 'N': 0.039360537265493829,  \
   'Q': 0.044689393386378569, 'P': 0.043652821373822905, 'S': 0.058697715161690638, \
   'R': 0.056001167968464854, 'T': 0.053984962406015038, 'W': 0.01527264763851376, \
   'V': 0.070679611650485433, 'Y': 0.028486750857726843}

  # rand dict maps int:aa
  
  # massLeft = {}
  # for i in range(MIN_MASS,MAX_MASS): massLeft[i] = True

  # maps mass:[peptide]
  mass2seq = {}
  
  # keep generating peptides until we have no masses left between MAX_MASS
  # and MIN_MASS
  # MULTI possibly take out the R and K?
  rFreq = aaFrac.pop('R')
  kFreq = aaFrac.pop('K')

  randDict = frequencyToRandDict(aaFrac)

  while len(massLeft.keys()) > 0:
    peptide, massInt = generateTrypticPeptide(\
                    randDict, kFreq=kFreq, rFreq=rFreq, massesAllowed=massLeft)
    # if ( (mass > MAX_MASS) or (mass < MIN_MASS) ): continue

    #if len(massLeft) % 100 == 0: 
    #  print >>sys.stderr, "Only %i masses left" % len(massLeft)
    # print "peptide = ", peptide
    # print "massLeft = ", len(massLeft)
    # print "-----"
    if not massLeft.has_key(massInt): continue

    # add to mass2seq
    if mass2seq.has_key(massInt): mass2seq[massInt].append(peptide) 
    else: mass2seq[massInt] = [ peptide ]

    # remove this mass if we have enough peptides
    if massLeft.has_key(massInt):
      if len(mass2seq[massInt]) >= MIN_PEPTIDES: 
        massLeft.pop(massInt)

  return mass2seq

#-------------------------------------------------------------------------
def setRanges(spectrum, peptideRunner):
  # assumed m is M+H+
  mass = float(spectrum["assumed m"]) - Mass.mono_mass('H_plus') 
  maxMass = mass + Mass.tol_mass('prec')
  minMass = mass - Mass.tol_mass('prec')
  peptideRunner.min_mass = minMass
  peptideRunner.max_mass = maxMass

#------------------------------------------------------------------------
def trulyRandomNegative(fastaFname=None, spectrumObjectFname=None):
  fastaPrefix = os.path.basename(fastaFname).split('.')[0]
  objPrefix = os.path.basename(spectrumObjectFname).split('.')[0] 
  randomObs = objPrefix  + ",trurand.pickledpsm"

  # don't redo if not necessary
  if os.path.isfile(randomObs) and not ALL_REDO:
    if os.path.getsize(randomObs) > 0:
      print >>sys.stderr, "Using old %s" % randomObs
      return randomObs
    else:
      print >>sys.stderr, "Ignoring zero size %s" % randomObs

  # all masses are real masses. that is, masses of the charged peptide
  suid2mass = {}
  print >>sys.stderr, "Loading PsmCollection %s" % spectrumObjectFname
  psmColl = cPickle.load(open(spectrumObjectFname))
  for spectrum in psmColl.allSpectra():
    suid2mass[spectrum.uid()] = int(float(spectrum["assumed m"]) + Mass.mono_mass('H_plus') + 0.5)

  massesAllowed = {}
  for value in suid2mass.values(): 
    for mass in range(value-MASS_WINDOW, value + MASS_WINDOW + 1):
      massesAllowed[mass] = 1

  print >>sys.stderr, "Creating null peptides for %s" % fastaPrefix

  nullDict = nullPeptideDict(fastaFname, massesAllowed)
  nullPickle = open(fastaPrefix + ".nulldict", "w")
  cPickle.dump(nullDict, nullPickle)
  nullPickle.close()

  for match in psmColl["matches"]:
    suid = match.spectrum.uid()
    nullPeps = nullDict[suid2mass[suid]]
    randompep = nullPeps[ int( random.random() * len(nullPeps) )]
    # print randompep
    match.peptide = randompep

  cPickle.dump(psmColl, open(randomObs, "w"))
  return randomObs

#------------------------------------------------------------------------
def candidateNegative(collectionFile=None, fastaFname=None, candidatePeptideNumber=CANDIDATE_PEPTIDE_NUMBER):

  newMatchCollectionName = os.path.basename(collectionFile).split('.')[0] \
      + ",cand.pickledpsm"

  # don't redo if not necessary
  if os.path.isfile(newMatchCollectionName) and not ALL_REDO:
    if os.path.getsize(newMatchCollectionName) > 0:
      print >>sys.stderr, "Using old %s" % newMatchCollectionName
      return newMatchCollectionName
    else:
      print >>sys.stderr, "Ignoring zero size %s" % newMatchCollectionName 

  # parse the spectrum collection
  coll = cPickle.load(open(collectionFile))

  # create the new match collection
  newColl = MatchCollection.PsmMatchCollection()

  for match in coll["matches"]:
    match.peptide.replace("I", "L")
    suid = str(match.spectrum.uid())

    spectrum = match.spectrum
    spectrum.parsePeaks()
    peptideRunner = Crux.GeneratePeptideRunner(fastaFname)
    setRanges(spectrum, peptideRunner) 
    peptideRunner.use_index = "F"
    peptideRunner.run()
    peps = peptideRunner.getPeptides()

    if len(peps) == 0: 
      raise SystemExit, "Problem with match %s, no peptides!" % suid
    
    random.shuffle(peps)
    pepCount = 0
    for pep in peps:
      pepCount += 1
      pep = pep.replace("I", "L")
      if (pep != match.peptide) and (pepCount <= candidatePeptideNumber):
        newMatch = Match.Match()
        # TODO create a copy match function to do the following
        newMatch.spectrum = match.spectrum
        newMatch.peptide = pep
        newMatch.charge = match.charge
        newColl.appendMatch(newMatch)

  cPickle.dump(newColl, open(newMatchCollectionName, 'w'))
  return newMatchCollectionName


#------------------------------------------------------------------------
def outputPeps(suid2mass, dict, outDir):
  """
  I suid2mass maps suid:massint 
  I dict maps massint:[peps]
  I outDir is dir to place pep files
  O suid2pepFiles maps suid:file
  """
  suid2pepFiles = {}
  for suid in suid2mass:
    massInt = suid2mass[suid]
    outFilename = outDir + "/" + suid + ".cand"
    outFile = open(outFilename, "w")
    keptPeps = []
    for curMassInt in range(massInt-MASS_WINDOW, massInt+MASS_WINDOW + 1):
      if not dict.has_key(curMassInt): continue
      peps = dict[curMassInt]
      for pep in peps: keptPeps.append(pep)

    for pep in keptPeps: outFile.write(str(pep) + "\n")
    outFile.close()
    suid2pepFiles[suid] = outFilename
  return suid2pepFiles
                
#------------------------------------------------------------------------
def outputPepFiles(prefix, candFiles):
  """
  candFiles maps uid:candfile
  prefix is the prefix for the <prefix>.peps file
  """
  uidkeys = candFiles.keys()
  fh = open(prefix + ".peps", 'w')
  for uid in uidkeys:
    string = "%s\t%s\n" % (uid, candFiles[uid])
    fh.write(string)
  fh.close()

#------------------------------------------------------------------------
def getMassesAllowed(candDict, suid2mass):
  possibleMasses = {}
  for mass in candDict.keys(): possibleMasses[mass] = 1
  massesAllowed = {}
  for mass in suid2mass.values(): 
    for massRange in range(mass-MASS_WINDOW, mass+MASS_WINDOW):
      if massRange > MAX_MASS: continue
      if massRange < MIN_MASS: continue
      if not possibleMasses.has_key(massRange): 
        print >>sys.stderr, "Mass %i not possible" % massRange
        continue
      massesAllowed[massRange] = 1
  return massesAllowed

#-----------------------------------------
# MAIN
#-----------------------------------------
# returns the pepFname
#-----------------------------------------
def main(fastaFname=None, spectrumObjectFname=None):
  fastaPrefix = os.path.basename(fastaFname).split('.')[0]
  objPrefix = os.path.basename(spectrumObjectFname).split('.')[0] 
  joinedPrefix = os.getcwd() + os.sep + os.path.basename(objPrefix) + "," + \
      os.path.basename(fastaPrefix) 

  pepsName = joinedPrefix + ".peps"
  # Avoid rerunning 
  if os.path.isfile(pepsName) and not ALL_REDO:
    if os.path.getsize(pepsName) > 0:
      print >>sys.stderr, "Using old %s" % pepsName
      return pepsName
    else:
      print >>sys.stderr, "Ignoring zero size %s" % pepsName
  
  # all masses are real masses. that is, masses of the charged peptide
  suid2mass = {}
  if ".pickledscoll" in spectrumObjectFname:
    print >>sys.stderr, "Loading SpectrumCollection %s" % spectrumObjectFname
    specColl = cPickle.load(open(spectrumObjectFname))
    for spectrum in specColl.allSpectra():
      suid2mass[spectrum.uid()] = spectrum.integerMass()
    del specColl
  elif ".pickledpsm" in spectrumObjectFname:
    print >>sys.stderr, "Loading PsmCollection %s" % spectrumObjectFname
    psmColl = cPickle.load(open(spectrumObjectFname))
    for spectrum in psmColl.allSpectra():
      suid2mass[spectrum.uid()] = spectrum.integerMass()
    del psmColl
  else:
    raise SystemExit, "Need a .pickledscoll or a .pickledpsm!"

  massesAllowed = {}
  for value in suid2mass.values(): 
    for mass in range(value-MASS_WINDOW, value + MASS_WINDOW + 1):
      massesAllowed[mass] = 1

  print >>sys.stderr, "Creating candidate peptides for %s" % fastaFname
  candDict = candTrypticPeptideDict(fastaPrefix, massesAllowed)

  print >>sys.stderr, "Outputting candidate peptides for %s against %s" % \
    (spectrumObjectFname, fastaFname) 
  outDir = joinedPrefix + ".pepdir"
  if not os.path.isdir(outDir): os.mkdir(outDir)
  suid2candPepFiles = outputPeps(suid2mass, candDict, outDir)

  print >>sys.stderr, "Outputting pepfile names to %s.peps" % joinedPrefix
  outputPepFiles(joinedPrefix, suid2candPepFiles)
  return pepsName

#-----------------------------------------
# MAIN
#-----------------------------------------
if __name__ == "__main__":
  if len(sys.argv) < 3 :
    raise SystemExit, "Usage: %s <fasta-filename> <psm-filename>" % sys.argv[0]
  posPsmCollectionName = sys.argv[1]
  fastaFname = sys.argv[2]
  candidatePeptideNumbers = [10000]
  for candidatePeptideNumber in candidatePeptideNumbers:
    print >>sys.stderr, "Creating PsmCollection with %i peptides." % candidatePeptideNumber
    candidatePsmCollectionName = candidateNegative(\
      collectionFile=posPsmCollectionName, fastaFname=fastaFname, \
      candidatePeptideNumber=candidatePeptideNumber)
    newName = candidatePsmCollectionName.replace(".", "-%i." % candidatePeptideNumber)
    print "New name = %s" % newName
    os.rename(candidatePsmCollectionName, newName)
