#---------------------------------------------------------------------------
# Name: generatePeptides.py
#---------------------------------------------------------------------------
# Author: $Author: aklammer $
#---------------------------------------------------------------------------
# Revision: $Revision: 1.3 $
#---------------------------------------------------------------------------
# Inputs:
#  - a <prefix>.fasta file from which to extract aa frequencies
#  - a cPickled SpectrumCollection with spectra 
#    for which to generate candidate and null peptides 
#--------------------------------------------------------------------------
# Outputs:
#  - a directory <prefix>.pepdir containing two files for each spectrum:
#      - spectrum-id.null - synthetic peptides for the null distribution 
#      - spectrum-id.cand - candidate peptides for matches
#  - an output file <prefix>.peps mapping the spectra to their peptide files: 
# 
#     spectrum id \t null peps file \t cand peps file \n
#---------------------------------------------------------------------------
# NOTE: This is meant to faithfully represent PepHMM. For a real database
# search, you would want to have peptides near the same m/z (I think) not the
# same mass, since it is the m/z that is known. In PepHMM, we deal only with
# +2 spectra, so this is not an issue. 
#---------------------------------------------------------------------------

# MIN_MASS = 400
MIN_MASS = 50
# MAX_MASS = 5000
MAX_MASS = 5000
# MIN_PEPTIDES = 1
MIN_PEPTIDES = 200
MIN_PEPTIDE_LEN = 3 
ITERATIONS = 4
MASS_WINDOW = 2
MAX_PEPTIDES = 500


import os, sys
if __name__ == "__main__":
  if len(sys.argv) < 2:
    print >>sys.stderr, """
Usage: %s <seq db prefix> <pickled object> 
    <seq db> - A <prefix>.fasta file for generating null and candidate peptides
    <pickled object> 
             - The pickled SpectrumCollection (with suffix .pickledscoll) or 
               PsmMatchCollection (with suffix .pickledpsm) object that 
               will be searched.
             
""" % os.path.basename(sys.argv[0])
    sys.exit(1)

try:
  import Bio.Fasta
except:
  print >>sys.stderr, "Could not import Bio.Fasta"

import random
import Mass
import SpectrumCollection, Spectrum
import cPickle

def candPeptideDict(seqs):
  """
  return a dict mapping mass:[peptides] 
  """

  # find tryptic seqs
  z = 2
  mass2seq = {}
  for seq in seqs:
    for start in range(len(seq)):
      mass = Mass.mono_mass('H2O') + Mass.mono_mass('H_plus') * z
      for end in range(start, len(seq)):
        mass += Mass.mono_mass(seq[end])
        massInt = int(mass)
        if (massInt < MIN_MASS): continue
        if (MAX_MASS < massInt): break
        if mass2seq.has_key(mass):
          mass2seq[massInt].append(seq[start:end+1])
        else:
          mass2seq[massInt] = [ seq[start:end+1] ]

  return mass2seq


def candTrypticPeptideDict(fastaPrefix, massesAllowed):
  """
  return a dict mapping mass:[peptides] 
  uses only tryptic peptides
  """

  print >>sys.stderr, "Reading fasta file %s" % fastaPrefix
  fh = open(fastaPrefix + ".fasta")
  iterator = Bio.Fasta.Iterator(fh, Bio.Fasta.RecordParser())
  idx = 0
  trypticSeqs = []
  while(True):
    idx += 1
    if idx % 10000 == 0: print >>sys.stderr, "Reached sequence %i" % idx
    curRecord = iterator.next()
    if curRecord is None: break
    seq = curRecord.sequence
    trypticSeq = []
    for i in range(len(seq)):
      trypticSeq.append(seq[i])
      if (seq[i] == 'K' or seq[i] == 'R'):
        if i < len(seq)-1:
          if seq[i+1] != 'P':

            if len(trypticSeq) >= MIN_PEPTIDE_LEN:
              trypticSequence = ''.join(trypticSeq)
              mass = Mass.peptide_mass(trypticSequence, z=2)
              massInt = int(mass + 0.5)
              if ( (massInt < MIN_MASS) or (MAX_MASS < massInt) or not \
                  massesAllowed.has_key(massInt)  or not \
                  Mass.canonicalOnly(trypticSeq)): 
                trypticSeq = []
                continue
              trypticSeqs.append(trypticSequence)
            trypticSeq = []

    # add any remaining seq
    if len(trypticSeq) >= MIN_PEPTIDE_LEN:
      trypticSequence = ''.join(trypticSeq)
      mass = Mass.peptide_mass(trypticSequence, z=2)
      massInt = int(mass + 0.5)
      if ( (massInt < MIN_MASS) or (MAX_MASS < massInt) or not \
          massesAllowed.has_key(massInt) or not \
          Mass.canonicalOnly(trypticSeq)): 
        trypticSeq = []
        continue
      trypticSeqs.append(trypticSequence)

  fh.close()
  # calc mass dict
  # FIX eliminate X's and other non-standard aas
  mass2seq = {}
  for peptide in trypticSeqs:
    massBin = int(Mass.peptide_mass(peptide, z=2) + 0.5)
    if mass2seq.has_key(massBin):
      if not mass2seq[massBin].has_key(peptide):
        mass2seq[massBin][peptide] = 1
    else:
      mass2seq[massBin] = { peptide : 1 }

  return mass2seq

#----------------------
PRECISION = 100
def frequencyToRandDict(freqDict):
  """
  return a list of length PRECISION where values are the keys of input dict.
  Each key in input dict is present as a value in output dict proportional
  to its corresponding value in input freqDict.
  """

  idx = 0
  rand = []
  for key in freqDict:
    for subIdx in range(int(freqDict[key]*PRECISION)):
      rand.append(key)
      idx+=1
  if len(rand) < PRECISION: rand.append(rand[-1])
  return rand


def reverse(peptide):
  rev = ""
  for i in range(len(peptide)):
    rev+=peptide[len(peptide)-1-i]
  return rev

#--------------------------------------------------------------------
#--------------------------------------------------------------------
# generate a random tryptic 
# peptide with aa freq aaFreq within one of the masses in massesAllowed
#--------------------------------------------------------------------
def generateTrypticPeptide(randDict, kFreq=0.5, rFreq=0.5, massesAllowed=None, z=2):
  """
i aa:frequency 
i mass:exist
o peptide
  """
  kThresh = kFreq / (kFreq + rFreq)
   
  while(True):
    endingAA = ""
    if random.random() < kThresh: endingAA = 'K'
    else: endingAA = 'R'

    peptide = endingAA
    mass = Mass.peptide_mass(endingAA, z=2)

    for i in range(ITERATIONS):

      while mass < MAX_MASS:
        raa = randDict[ int(len(randDict)*random.random()) ] # Mass.randomAA()
        peptide += raa
        mass += Mass.mono_mass(raa)
        massInt = int(mass+0.5)
        if massesAllowed.has_key(massInt): 
          return reverse(peptide), massInt

      while MIN_MASS < mass and len(peptide) > 1:
        raa = peptide[-1]
        mass -= Mass.mono_mass(raa)
        peptide = peptide[0:len(peptide)-1] 

#-------------------------------------
def nullPeptideDict(fastaPrefix, massLeft):
  """
  return a dict mapping mass:[peptides] 
  uses only tryptic peptides
  """

  # count aa frequencies
  aaCount = {}
  for aa in Mass.CanonicalAminos: aaCount[aa] = 0
  totalCount = 0

  print >>sys.stderr, "Reading fasta file %s" % fastaPrefix
  fh = open(fastaPrefix + ".fasta")
  iterator = Bio.Fasta.Iterator(fh, Bio.Fasta.RecordParser())
  idx = 0
  while(True):
    idx += 1
    if idx % 10000 == 0: print >>sys.stderr, "Reached sequence %i" % idx
    curRecord = iterator.next()
    if curRecord is None: break
    seq = curRecord.sequence
    totalCount += len(seq)
    for idx in range(len(seq)): 
      if aaCount.has_key(seq[idx]): aaCount[seq[idx]]+=1

  fh.close()

  aaFrac = {} 
  for key in aaCount.keys(): aaFrac[key] = float(aaCount[key])/totalCount

  # rand dict maps int:aa
  
  # massLeft = {}
  # for i in range(MIN_MASS,MAX_MASS): massLeft[i] = True

  # maps mass:[peptide]
  mass2seq = {}
  
  # keep generating peptides until we have no masses left between MAX_MASS
  # and MIN_MASS
  rFreq = aaFrac.pop('R')
  kFreq = aaFrac.pop('K')

  randDict = frequencyToRandDict(aaFrac)

  while len(massLeft.keys()) > 0:
    peptide, massInt = generateTrypticPeptide(\
                    randDict, kFreq=kFreq, rFreq=rFreq, massesAllowed=massLeft)
    # if ( (mass > MAX_MASS) or (mass < MIN_MASS) ): continue

    if len(massLeft) % 100 == 0: 
      print >>sys.stderr, "Only %i masses left" % len(massLeft)
    # print "peptide = ", peptide
    # print "massLeft = ", len(massLeft)
    # print "-----"
    if not massLeft.has_key(massInt): continue

    # add to mass2seq
    if mass2seq.has_key(massInt): mass2seq[massInt].append(peptide) 
    else: mass2seq[massInt] = [ peptide ]

    # remove this mass if we have enough peptides
    if massLeft.has_key(massInt):
      if len(mass2seq[massInt]) >= MIN_PEPTIDES: 
        massLeft.pop(massInt)

  return mass2seq



def outputPeps(suid2mass, dict, outDir, suffix):
  """
  I suid2mass maps suid:massint 
  I dict maps massint:[peps]
  I outDir is dir to place pep files
  I suffix is whether we are .cand or .null
  O suid2pepFiles maps suid:file
  """
  suid2pepFiles = {}
  for suid in suid2mass:
    massInt = suid2mass[suid]
    outFilename = outDir + "/" + suid + suffix
    outFile = open(outFilename, "w")
    keptPeps = []
    for curMassInt in range(massInt-MASS_WINDOW, massInt+MASS_WINDOW + 1):
      if not dict.has_key(curMassInt): continue
      peps = dict[curMassInt]
      for pep in peps: keptPeps.append(pep)
    if suffix == ".null":
      # reduce keptPeps to 500 only
      keptPeps.sort(lambda x,y: cmp(random.random(), random.random()))
      keptPeps = keptPeps[0:500]

    for pep in keptPeps: outFile.write(str(pep) + "\n")
    outFile.close()
    suid2pepFiles[suid] = outFilename
  return suid2pepFiles
                

def outputPepFiles(prefix, nullFiles, candFiles):
  """
  nullFiles maps uid:nullfile
  candFiles maps uid:candfile
  prefix is the prefix for the <prefix>.peps file
  """
  uidkeys = nullFiles.keys()
  fh = open(prefix + ".peps", 'w')
  for uid in uidkeys:
    string = "%s\t%s\t%s\n" % (uid, nullFiles[uid], candFiles[uid])
    fh.write(string)
  fh.close()

def getMassesAllowed(candDict, suid2mass):
  possibleMasses = {}
  for mass in candDict.keys(): possibleMasses[mass] = 1
  massesAllowed = {}
  for mass in suid2mass.values(): 
    for massRange in range(mass-MASS_WINDOW, mass+MASS_WINDOW):
      if massRange > MAX_MASS: continue
      if massRange < MIN_MASS: continue
      if not possibleMasses.has_key(massRange): 
        print >>sys.stderr, "Mass %i not possible" % massRange
        continue
      massesAllowed[massRange] = 1
  return massesAllowed


#-----------------------------------------
# MAIN
#-----------------------------------------
if __name__ == "__main__":

  fastaPrefix = sys.argv[1]
  objPrefix = sys.argv[2].split('.')[0]
  joinedPrefix = os.path.basename(objPrefix) + "," + \
                  os.path.basename(fastaPrefix) 
  
  # all masses are real masses. that is, masses of the charged peptide
  suid2mass = {}
  if ".pickledscoll" in sys.argv[2]:
    print >>sys.stderr, "Loading SpectrumCollection %s" % sys.argv[2]
    specColl = cPickle.load(open(sys.argv[2]))
    for spectrum in specColl.allSpectra():
      suid2mass[spectrum.uid()] = int(float(spectrum["assumed m"]) + Mass.mono_mass('H_plus')  + 0.5)
    del specColl
  elif ".pickledpsm" in sys.argv[2]:
    print >>sys.stderr, "Loading PsmCollection %s" % sys.argv[2]
    psmColl = cPickle.load(open(sys.argv[2]))
    for spectrum in psmColl.allSpectra():
      suid2mass[spectrum.uid()] = int(float(spectrum["assumed m"]) + Mass.mono_mass('H_plus') + 0.5)
    del psmColl
  else:
    raise SystemExit, "Need a .pickledscoll or a .pickledpsm!"

  massesAllowed = {}
  for value in suid2mass.values(): 
    for mass in range(value-MASS_WINDOW, value + MASS_WINDOW + 1):
      massesAllowed[mass] = 1

  print >>sys.stderr, "Creating candidate peptides for %s" % fastaPrefix
  candDict = candTrypticPeptideDict(fastaPrefix, massesAllowed)
  candPickle = open(fastaPrefix + ".canddict", "w")
  cPickle.dump(candDict, candPickle)
  candPickle.close()

  print >>sys.stderr, "Outputting candidate peptides for %s against %s" % \
    (objPrefix, fastaPrefix) 
  outDir = joinedPrefix + ".pepdir"
  if not os.path.isdir(outDir): os.mkdir(outDir)
  suffix = ".cand"
  suid2candPepFiles = outputPeps(suid2mass, candDict, outDir, suffix)

  massesAllowed = getMassesAllowed(candDict, suid2mass)
  del candDict
  
  print >>sys.stderr, "Creating null peptides for %s" % fastaPrefix
  nullDict = nullPeptideDict(fastaPrefix, massesAllowed)
  nullPickle = open(fastaPrefix + ".nulldict", "w")
  cPickle.dump(nullDict, nullPickle)
  nullPickle.close()
  
  print >>sys.stderr, "Outputting null peptides for %s against %s" % \
    (objPrefix, fastaPrefix)
  suffix = ".null"
  suid2nullPepFiles = outputPeps(suid2mass, nullDict, outDir, suffix)

  print >>sys.stderr, "Outputting pepfile names to %s.peps" % joinedPrefix
  outputPepFiles(joinedPrefix, suid2nullPepFiles, suid2candPepFiles)
