# ---------------------------------------------------------------------------
# File: hmmTestSpectrumSion.py
#---------------------------------------------------------------------------
# Author: $Author: aklammer $
#---------------------------------------------------------------------------
# Revision: $Revision: 1.28 $
#---------------------------------------------------------------------------
# Inputs:
# 
#  - a spectrum_uid 
#  - <prefix>.peps, a file containing the candidate peptides
#  - a <collection-prefix>.scoll|pickledpsm which contains the spectrum
#  - a trained GMTK model specifier  
#  - a trained GMTK model params file specifier
#  - optional working root directory. cwd if not given
#---------------------------------------------------------------------------
# Outputs:
#
#  - a file <prefix>.scores, containing the lors for each ion type
#    for each peptide-spectrum match:
# 
#     spectrum id;peptide \t lor-1 \t ...  \t lor-n \n
# 
#---------------------------------------------------------------------------
import sys, os.path
import Crux
import generatePeptides
import GMTKmodel

#---------------------------------------------------------------------------
if __name__ == "__main__":
  if len(sys.argv) < 6:
    print >>sys.stderr, """
Usage: %s <spectrum-uid> <collection-file> <fasta-file> <GMTK-model-name> <GMTK-params-file> <svm-prefix> <is-null?>
   <spectrum-uid>      - spectrum universal identifier (e.g. 10000,E_coli,2)
   <collection-file>   - a pickled file (.pickledscoll | .pickledpsm) 
                         containing the spectrum
   <indexed-fasta-file> - an indexed fasta file
   <GMTK-model-name>   - trained GMTK model specifier  
   <GMTK-params-files> - a trained GMTK model params file specifier
   <svm-prefix> - Svm prefix
   <null> - (T|F)

   output probs and zscores to a <spectrum-uid>.scores file
  """ % os.path.basename(sys.argv[0])
    sys.exit(1)
 
#------------------------------------------------------------------------
import sys, random, cPickle, tempfile, shutil, signal
import Mass
import math
import Numeric
import Match
import MatchCollection
import Spectrum
import Sp
from GMTK import *
from hmmParams import *
from hmmAccuracySionSvm import *

#------------------------------------------------------------------------
global TEMP_DIR
TEMP_DIR = None

#------------------------------------------------------------------------
# this code will be designed to run on the cluster. In other words it
# will create it's own working dir, copy over the specified input 
# files to this directory, then read in the working files, process them
# and then copy the output files back to it's original working directory.
# if at any point the processing fails, it will remove it's working
# directory, and write to STDERR about what happened
#------------------------------------------------------------------------

#-------------------------------------------------------------------------
def sigtermHandler(a,b): raise Exception("Got a SIGTERM!")

#-------------------------------------------------------------------------
def abort(message):
  global TEMP_DIR
  if CLEANUP != None:
    try: 
      os.chdir('/tmp')
      if CLEANUP==True: shutil.rmtree(TEMP_DIR)
    except Exception, e: 
      message += " - Also, could not delete %s because %s" % (TEMP_DIR,str(e))
  print >>sys.stderr, "FATAL - %s %s %s - %s" % \
        (sys.argv[1],sys.argv[2],sys.argv[3],message)
  raise SystemExit, -1

#-------------------------------------------------------------------------
def parsePeps(candFile):
  peps = []
  fh = open(candFile)
  for line in fh:
    peptide = line.rstrip('\n')
    if not Mass.canonicalOnly(peptide): 
      print >>sys.stderr, "Skipping non-canonical aa at idx %s" % peptide
      continue
    peps.append(peptide)
  fh.close()
  return peps


#-------------------------------------------------------------------------
def main(argv):
  spectrumUid = argv[1]
  collectionFile = argv[2]
  fastaFile = argv[3]
  modelName = argv[4]
  gmtkParms = argv[5]
  svmPrefix = argv[6]
  isNull = argv[7]
  cleavages = CLEAVAGES
  print >>sys.stderr, "Cleavages: %s" % CLEAVAGES

  model = GMTKmodel.gmtkModelFactory(modelName)
  if isNull == "T": nullString = "null"
  elif isNull == "F": nullString = ""
  else:  raise SystemExit, "Strange null string %s" % nullString

  # create temp directory within workingDir
  suffix = '.hmmtestdir'
  prefix = spectrumUid + '-'
  originalDir = os.getcwd()
  global TEMP_DIR
  try: 
    TEMP_DIR = tempfile.mkdtemp(suffix, prefix, WORKING_DIR)
  except: 
    raise Exception, "Failed to make TEMP_DIR in directory '%s'" % WORKING_DIR

  # copy all input files to TEMP_DIR, report any error
  try:
    shutil.copy(collectionFile, TEMP_DIR)
  except: 
    raise Exception, "Failed to copy files to temp directory '%s'" % TEMP_DIR

  # cd to TEMP_DIR
  try: 
    os.chdir(TEMP_DIR)
  except: 
    raise Exception, "Failed to cd to directory %s!" % TEMP_DIR

  # begin processing of files, report any error
  match = None
  try:
    coll = cPickle.load(open(collectionFile)) # might be psmColl or specColl
    match = coll.getMatch(spectrumUid)
    spectrum = match.spectrum
    spectrum.parsePeaks()
    del coll
    peptideRunner = Crux.GeneratePeptideRunner(fastaFile)
    generatePeptides.setRanges(spectrum, peptideRunner) 
    peptideRunner.cleavages = cleavages
    peptideRunner.run()
    peps = peptideRunner.getPeptides()
    # peps = ['EAMAPK']
    print >>sys.stderr, "Got %i peps!" % len(peps)
  except Exception, val: 
    raise Exception, "Failed to get peptides for spectrum: %s" % str(val)

  try:
    analyzePeps = []
    if isNull == 'F':
      filteredPeps = Sp.spFilter(spectrum, peps)
      seenTruePep = False
      for pep in filteredPeps: 
        if pep == match.peptide: seenTruePep = True
      if seenTruePep == False: 
        if match.peptide != '' and isNull == 'F':
          print >>sys.stderr, "Forced to append true pep to candidate peps for %s" \
            % match.spectrum.uid()
          filteredPeps += [ match.peptide ] 
      analyzePeps = filteredPeps
      # analyzePeps = ['EAMAPK']
    else:
      random.shuffle(peps) 
      analyzePeps = peps[0:MAX_PEPS]
    psmCollection = MatchCollection.PsmMatchCollection()
    psmCollection.addPsmsForSpectrum(spectrum, analyzePeps)
  except Exception, val: 
    raise Exception, "Failed to prepare spectrum: %s" % str(val)

  try:
    psmCollection["file"] = "testPsmColl" + nullString # the routines below need this
    if KERNEL == 'lor':
      scores = sionLor(psmCollection, model, gmtkParms)
    elif KERNEL == 'accumulator' or KERNEL == 'fisher':
      scores = sionFisherKernel(psmCollection, model, gmtkParms)
  except Exception, val: 
    raise Exception, "Failed to process spectrum: %s" % str(val)

  # the accumulator kernel is too big to be output locally
  svmInputFile = spectrumUid + "." + nullString + "svminput"
  if KERNEL == 'lor': svmInputFile = originalDir + os.sep + svmInputFile 

  try: 
    fh = open(svmInputFile, 'w'); fh.write(scores.svmString()); fh.close();
  except Exception, e: 
    raise Exception, "Could not output svminput %s" % str(e)

  # run SVM
  try:
    pep2score = singleSvmAccuracy(svmInputFile, svmPrefix)
  except Exception, e:
    raise Exception, "Could not run svm %s" % str(e)

  # output scores
  try: 
    fh = open(originalDir + os.sep + spectrumUid + "." + nullString + "scores", 'w')
    for pep in pep2score.keys(): fh.write("%s\t%s\n" % (pep, pep2score[pep]))
    fh.close()
  except Exception, e: 
    raise Exception, "Could not output scores! %s" % str(e)

  os.chdir(WORKING_DIR)
  if CLEANUP == True:
    try: shutil.rmtree(TEMP_DIR)
    except: raise Exception, "Failed to remove TEMP_DIR %s!" % TEMP_DIR

#-------------------------------------------------------------------------
if __name__ == "__main__":
  signal.signal(signal.SIGTERM, sigtermHandler)
  try: main(sys.argv)
  except Exception, e: abort(str(e))

  print >>sys.stderr, "SUCCESS - %s %s %s" % \
    (sys.argv[1], sys.argv[2], sys.argv[3])

  sys.exit(0)
