#---------------------------------------------------------------------------
# File: retentionRoc.py
#---------------------------------------------------------------------------
# Author: $Author: aklammer $
#---------------------------------------------------------------------------
# Revision: $Revision: 1.2 $
#---------------------------------------------------------------------------
# Inputs:
# 
#  - a <collection-file> which contains the spectra 
#  - the output of generatePeptides.py, <prefix>.peps, which maps spectra 
#    to candidate peptides:
#
#     SPECTRUM_ID \t HMM_CAND_FILE \n
#     - SPECTRUM_ID is the identifier of the spectrum
#     - HMM_CAND_FILE is the file containing the candidate peptides
#---------------------------------------------------------------------------
# Outputs:
#
#  - a single file with a row for each spectrum in <collection-file>
#    which has the suid, its positive peptide and its pivot peptide (the
#    previous peptide, when sorted by scan number)
#  - a file for each candidate peptide <prefix>.rpscore
#    which has the distance of each candidate peptide from its spectrum's 
#    pivot peptide
#  - a file, appropriate for roc analysis, with "+1 <\t> score <\n>" for all 
#    positive peptides and "-1 <\t> score <\n>" for all negative peptides
#---------------------------------------------------------------------------
import os, sys

#---------------------------------------------------------------------------
if __name__ == "__main__":
  if len(sys.argv) < 3:
    print >>sys.stderr, """
Usage: %s <collection-file> <peps-file>
    <collection-file> - a pickledpsm file
    <peps> - The .peps file.
  """ % os.path.basename(sys.argv[0])
    sys.exit(1)

#---------------------------------------------------------------------------
import os.path, pickle, tempfile
import Mass
import HmmData
import math
import Numeric
import MatchCollection


#---------------------------------------------------------------------------
aminoPositionScoreDict = { 'W':-4.0, 'F':-7.0, 'L':-9.0, 'I':-8.0, 'M':-5.5, \
  'V':-5.5, 'Y':-3.0, 'A':-1.5, 'T':5.0, 'P':4.0, 'E':7.0, 'D':9.0, 'C':4.0,\
  'S':5.0, 'Q':1.0, 'G':5.0, 'N':5.0, 'R':8.0, 'H':4.0, 'K':4.6}

#---------------------------------------------------------------------------
positionCoefficients = [ 0.42, 0.22, 0.05 ]

#---------------------------------------------------------------------------
aminoScoreDict = { 'W':11.0, 'F':10.5, 'L':9.6, 'I':8.4, 'M':5.8, 'V':5.0, \
  'Y':4.0, 'A':0.8, 'T':0.4, 'P':0.2, 'E':0.0, 'D':-0.5, 'C':-0.8, 'S':-0.8, \
  'Q':-0.9, 'G':-0.9, 'N':-1.2, 'R':-1.3, 'H':-1.3, 'K':-1.9}

#---------------------------------------------------------------------------
def retentionScore(bait, usePosition=True, useLength=True):
  score = 0.0
  for idx in range(len(bait)): score += aminoScoreDict[bait[idx]]

  if usePosition == True:
    for idx in range(0,3):
      score += positionCoefficients[idx] * aminoPositionScoreDict[bait[idx]]

  Kl = 0.0
  N = len(bait)
  if N < 10:
    Kl = 1 - 0.027 * (10 - N)
  elif N > 20:
    Kl = 1 - 0.014 * (N-20)
  else:
    Kl = 1.0

  H = score

  if H < 38:
    H = Kl * H
  else:
    H = Kl * H - 0.3 * (Kl*H - 38)

  return H
  
#---------------------------------------------------------------------------
USE_POSITION=True
USE_TWO=True
USE_SCAN_DIST=True
global retentionScoreDict
retentionScoreDict = {}

#---------------------------------------------------------------------------
def retentionDistance(bait, prey):

  global retentionScoreDict
  baitRT = preyRT = None

  if retentionScoreDict.has_key(bait):
    baitRT = retentionScoreDict[bait]
  else:
    baitRT = retentionScore(bait, usePosition=USE_POSITION)
    retentionScoreDict[bait] = baitRT

  if retentionScoreDict.has_key(prey):
    preyRT = retentionScoreDict[prey]
  else:
    preyRT = retentionScore(prey, usePosition=USE_POSITION)
    retentionScoreDict[prey] = preyRT

  return abs(baitRT - preyRT)

#---------------------------------------------------------------------------
# Returns nothing
#---------------------------------------------------------------------------
def main(collectionFile=None, pepsFile=None):
  #----- parse the .peps file
  suid2cand = {}
  pepsFh = open(pepsFile)
  for line in pepsFh:
    line = line.rstrip('\n')
    suid, candFile = line.split()
    fh = open(candFile)
    suid2cand[suid] = []
    for line in fh:
      line = line.rstrip('\n')
      suid2cand[suid].append(line)
    fh.close()
  pepsFh.close()
  #----- parse the spectrum collection
  coll = pickle.load(open(collectionFile))
  #----- sort the spectra by scan number
  matches = coll["matches"]
  matches.sort(lambda x,y: cmp(x.spectrum["first scan"], \
      y.spectrum["first scan"]))
  #----- construct file names
  posOutputFname = os.path.basename(collectionFile).split('.')[0] + ".preydist"
  negOutputFname = os.path.basename(collectionFile).split('.')[0] + ".chumdist"
  rocOutputFname = os.path.basename(collectionFile).split('.')[0] + ".rocinput"
  #----- open files
  posOutputFh = open(posOutputFname, 'w')
  negOutputFh = open(negOutputFname, 'w')
  rocOutputFh = open(rocOutputFname, 'w')
  #----- examine each consecutive pair of spectra. 
  for idx in range(len(matches)-2):
    #----- the first in a pair is the bait. 
    bait  = matches[idx].peptide
    bait2 = matches[idx+1].peptide
    #----- the second (or second and third) is prey. 
    prey = matches[idx+2].peptide
    #----- use the hydrophobicity of the bait to retrieve the prey.
    distance1 = retentionDistance(bait, prey)
    distance2 = retentionDistance(bait2, prey)
    #----- use the distance between scans as a proxy for rtime
    if USE_SCAN_DIST == True:
      distance1 /= float(matches[idx+2].spectrum["first scan"] \
          - matches[idx+0].spectrum["first scan"])
      distance2 /= float(matches[idx+2].spectrum["first scan"] \
          - matches[idx+1].spectrum["first scan"])
    #----- use the distance for both baits
    if USE_TWO == True: distance = (distance1 + distance2) / 2.0
    else: distance = distance1
    #----- print to a file its distance from bait 
    posOutputFh.write("%s\t%s\t%.3f\n" % (bait, prey, distance))
    rocOutputFh.write("+1\t%.3f\n" % distance)
    #----- also examine each candidate peptide (chum) in relation to bait
    for chum in suid2cand[suid]:
      #----- print to a file their distances from bait
      distance1 = retentionDistance(bait, chum)
      distance2 = retentionDistance(bait2, chum)
      #----- use the distance between scans as a proxy for rtime
      if USE_SCAN_DIST == True:
        distance1 /= float(matches[idx+2].spectrum["first scan"] \
            - matches[idx+0].spectrum["first scan"])
        distance2 /= float(matches[idx+2].spectrum["first scan"] \
            - matches[idx+1].spectrum["first scan"])
      #----- use the distance for both baits
      if USE_TWO == True: distance = (distance1 + distance2) / 2.0
      else: distance = distance1
      #----- print to a file its distance from bait 
      negOutputFh.write("%s\t%s\t%.3f\n" % (bait, chum, distance))
      rocOutputFh.write("-1\t%.3f\n" % distance)
  #----- clean up
  posOutputFh.close()
  negOutputFh.close()
  rocOutputFh.close()
  
#---------------------------------------------------------------------------
if __name__ == "__main__":
  main(collectionFile=sys.argv[1], pepsFile=sys.argv[2])



