#---------------------------------------------------------------------------
# File: hmmTestSpectrumSion.py
#---------------------------------------------------------------------------
# Author: $Author: aklammer $
#---------------------------------------------------------------------------
# Revision: $Revision: 1.2 $
#---------------------------------------------------------------------------
# Inputs:
# 
#  - a spectrum_uid 
#  - <prefix>.peps, a file containing the candidate peptides
#  - a <collection-prefix>.scoll|pickledpsm which contains the spectrum
#  - a trained GMTK model specifier  
#  - a trained GMTK model params file specifier
#  - optional working root directory. cwd if not given
#---------------------------------------------------------------------------
# Outputs:
#
#  - a file <prefix>.scores, containing the lors for each ion type
#    for each peptide-spectrum match:
# 
#     spectrum id;peptide \t lor-1 \t ...  \t lor-n \n
# 
#---------------------------------------------------------------------------
import sys, os

#---------------------------------------------------------------------------
if __name__ == "__main__":
  if len(sys.argv) < 4:
    print >>sys.stderr, """
Usage: %s <spectrum-uid> <collection-file> <indexed-fasta-file>
   <spectrum-uid>      - spectrum universal identifier (e.g. 10000,E_coli,2)
   <collection-file>   - a pickled file (.pickledscoll | .pickledpsm) 
                         containing the spectrum
   <indexed-fasta-file> - an indexed fasta file
   <param-prefix> - parameter prefix
   output p-values to a <spectrum-uid>.scores file
  """ % os.path.basename(sys.argv[0])
    sys.exit(1)

import os.path
import Crux
import GMTKmodel
import FixSvm
import GaussianMixture


#------------------------------------------------------------------------
import sys, cPickle, tempfile, shutil, signal
import Mass
import Match
import MatchCollection
import Spectrum
import Svm
from GMTK import *
from hmmParams import *

#------------------------------------------------------------------------
global TEMP_DIR
TEMP_DIR = None

#------------------------------------------------------------------------
# this code will be designed to run on the cluster. In other words it
# will create it's own working dir, copy over the specified input 
# files to this directory, then read in the working files, process them
# and then copy the output files back to it's original working directory.
# if at any point the processing fails, it will remove it's working
# directory, and write to STDERR about what happened
#------------------------------------------------------------------------

#-------------------------------------------------------------------------
def sigtermHandler(a,b): raise Exception("Got a SIGTERM!")

#-------------------------------------------------------------------------
def abort(message):
  global TEMP_DIR
  if CLEANUP != None:
    try: 
      os.chdir('/tmp')
      if CLEANUP==True: shutil.rmtree(TEMP_DIR)
    except Exception, e: 
      message += " - Also, could not delete %s because %s" % (TEMP_DIR,str(e))
  print >>sys.stderr, "FATAL - %s %s %s - %s" % \
        (sys.argv[1],sys.argv[2],sys.argv[3],message)
  raise SystemExit, -1

#-------------------------------------------------------------------------
def parsePeps(candFile):
  peps = []
  fh = open(candFile)
  for line in fh:
    peptide = line.rstrip('\n')
    if not Mass.canonicalOnly(peptide): 
      print >>sys.stderr, "Skipping non-canonical aa at idx %s" % peptide
      continue
    peps.append(peptide)
  fh.close()
  return peps

#-------------------------------------------------------------------------
def getPepToScore(svmInput, prefix):

  svmFileName = "%s.svmobject" % prefix
  svmObject = cPickle.load(open(svmFileName))
  fh = open(svmInput)
  pep2score = {}
  for line in fh:
    fields = line.rstrip("\n").split(",")
    peptide = fields[0].split(".")[1]
    vector = map(float, fields[2:])
    pep2score[peptide] = svmObject.calc_discriminant(vector)
  fh.close()
  return pep2score

#-------------------------------------------------------------------------
def main(argv):
  spectrumUid = argv[1]
  collectionFile = argv[2]
  fastaFile = argv[3]
  paramPrefix = argv[4]

  pairedParamPrefix = paramPrefix + "-pair"
  
  # create temp directory within workingDir
  suffix = '.hmmtestdir'
  prefix = spectrumUid + '-'
  originalDir = os.getcwd()
  global TEMP_DIR
  try: 
    TEMP_DIR = tempfile.mkdtemp(suffix, prefix, WORKING_DIR)
  except: 
    raise Exception, "Failed to make TEMP_DIR in directory '%s'" % WORKING_DIR

  # copy all input files to TEMP_DIR, report any error
  try:
    shutil.copy(collectionFile, TEMP_DIR)
  except: 
    raise Exception, "Failed to copy files to temp directory '%s'" % TEMP_DIR

  # cd to TEMP_DIR
  try: 
    os.chdir(TEMP_DIR)
  except: 
    raise Exception, "Failed to cd to directory %s!" % TEMP_DIR

  # begin processing of files, report any error
  match = None
  try:
    psmCollection = cPickle.load(open(collectionFile)) # might be psmColl or specColl
    match = psmCollection.getMatch(spectrumUid)
    spectrum = match.spectrum
    spectrum.parsePeaks()
    # downsample collection
    psmCollection["matches"] = [match]
    psmCollection["file"] = "testPsmColl" 
  except Exception, val: 
    raise Exception, "Failed to get peptides for spectrum: %s" % str(val)

  singleModel = GMTKmodel.gmtkModelFactory(SINGLE_MODEL_NAME, charge=match.charge, training=False, paramPrefix=paramPrefix)
  pairModel = GMTKmodel.gmtkModelFactory(PAIR_MODEL_NAME, charge=match.charge, paramPrefix=paramPrefix)

  try:
    print >>sys.stderr, "Generating single scores"
    singleScores = sionLor(psmCollection, singleModel, paramPrefix)
    print >>sys.stderr, "Generating pair scores"
    pairScores = sionLor(psmCollection, pairModel, pairedParamPrefix)
  except Exception, val: 
    raise Exception, "Failed to process spectrum: %s" % str(val)

  singleSvmInputFile   = spectrumUid + "-single." + "svminput"
  pairSvmInputFile     = spectrumUid + "-pair." + "svminput"
  mergedSvmInputFile   = spectrumUid + "-merged." + "svminput"

  try: 
    fh = open(singleSvmInputFile, 'w'); fh.write(singleScores.svmString()); fh.close();
    fh = open(pairSvmInputFile, 'w'); fh.write(pairScores.svmString()); fh.close();
    FixSvm.merge(pairSvmInputFile, singleSvmInputFile, mergedSvmInputFile)
    shutil.copy(mergedSvmInputFile, originalDir + os.sep + mergedSvmInputFile) 
    
  except Exception, e: 
    raise Exception, "Could not output svminput %s" % str(e)

  os.chdir(WORKING_DIR)
  if CLEANUP == True:
    try: shutil.rmtree(TEMP_DIR)
    except: raise Exception, "Failed to remove TEMP_DIR %s!" % TEMP_DIR

#-------------------------------------------------------------------------
if __name__ == "__main__":
  signal.signal(signal.SIGTERM, sigtermHandler)
  try: main(sys.argv)
  except Exception, e: abort(str(e))

  print >>sys.stderr, "SUCCESS - %s %s %s %s" % \
    (sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])

  sys.exit(0)
