#---------------------------------------------------------------------------
# File: hmmTestSpectrumSion.py
#---------------------------------------------------------------------------
# Author: $Author: aklammer $
#---------------------------------------------------------------------------
# Revision: $Revision: 1.25 $
#---------------------------------------------------------------------------
# Inputs:
# 
#  - a spectrum_uid 
#  - <prefix>.peps, a file containing the candidate peptides
#  - a <collection-prefix>.scoll|pickledpsm which contains the spectrum
#  - a trained GMTK model specifier  
#  - a trained GMTK model params file specifier
#  - optional working root directory. cwd if not given
#---------------------------------------------------------------------------
# Outputs:
#
#  - a file <prefix>.scores, containing the lors for each ion type
#    for each peptide-spectrum match:
# 
#     spectrum id;peptide \t lor-1 \t ...  \t lor-n \n
# 
#---------------------------------------------------------------------------
import sys, os

#---------------------------------------------------------------------------
if __name__ == "__main__":
  if len(sys.argv) < 4:
    print >>sys.stderr, """
Usage: %s <spectrum-uid> <ms2-file> <indexed-fasta-file> <param-prefix>
   <spectrum-uid>       - spectrum universal identifier (e.g. 10000,E_coli,2)
   <ms2-file>           - ms2 file
   <indexed-fasta-file> - an indexed fasta file
   <param-prefix>       - parameter prefix
   output p-values to a <spectrum-uid>.scores file
  """ % os.path.basename(sys.argv[0])
    sys.exit(1)

import os.path
import Crux
import GMTKmodel
import FixSvm
import GaussianMixture
import SpectrumCollection


#------------------------------------------------------------------------
import sys, cPickle, tempfile, shutil, signal
import Mass
import Match
import MatchCollection
import Spectrum
import Svm
from GMTK import *
from hmmParams import *

#------------------------------------------------------------------------
global TEMP_DIR
TEMP_DIR = None
DUPSCORE = "DUPSCORE"

#------------------------------------------------------------------------
# this code will be designed to run on the cluster. In other words it
# will create it's own working dir, copy over the specified input 
# files to this directory, then read in the working files, process them
# and then copy the output files back to it's original working directory.
# if at any point the processing fails, it will remove it's working
# directory, and write to STDERR about what happened
#------------------------------------------------------------------------

#-------------------------------------------------------------------------
def sigtermHandler(a,b): raise Exception("Got a SIGTERM!")

#-------------------------------------------------------------------------
def abort(message):
  global TEMP_DIR
  if CLEANUP != None:
    try: 
      os.chdir('/tmp')
      if CLEANUP==True: shutil.rmtree(TEMP_DIR)
    except Exception, e: 
      message += " - Also, could not delete %s because %s" % (TEMP_DIR,str(e))
  print >>sys.stderr, "FATAL - %s %s %s - %s" % \
        (sys.argv[1],sys.argv[2],sys.argv[3],message)
  raise SystemExit, -1

#-------------------------------------------------------------------------
def parsePeps(candFile):
  peps = []
  fh = open(candFile)
  for line in fh:
    peptide = line.rstrip('\n')
    if not Mass.canonicalOnly(peptide): 
      print >>sys.stderr, "Skipping non-canonical aa at idx %s" % peptide
      continue
    peps.append(peptide)
  fh.close()
  return peps

#-------------------------------------------------------------------------
def getPepToScore(svmInput, prefix):

  svmFileName = "%s.svmobject" % prefix
  svmObject = cPickle.load(open(svmFileName))
  fh = open(svmInput)
  pep2score = {}
  for line in fh:
    fields = line.rstrip("\n").split(",")
    peptide = fields[0].split("-")[-1]
    vector = map(float, fields[2:])
    pep2score[peptide] = svmObject.calc_discriminant(vector)
  fh.close()
  return pep2score

#-------------------------------------------------------------------------
def main(argv):
  spectrumUid = argv[1]
  ms2File = argv[2]
  fastaFile = argv[3]
  paramPrefix = argv[4]

  pairedParamPrefix = paramPrefix + "-pair"
  
  # create temp directory within workingDir
  suffix = '.hmmtestdir'
  prefix = spectrumUid + '-'

  scan = spectrumUid.split(",")[0]
  charge = spectrumUid.split(",")[2]

  originalDir = os.getcwd()
  global TEMP_DIR
  try: 
    TEMP_DIR = tempfile.mkdtemp(suffix, prefix, WORKING_DIR)
    os.chdir(TEMP_DIR)
  except: 
    raise Exception, "Failed to make or cd to TEMP_DIR in directory '%s'" % WORKING_DIR

  # begin processing of files, report any error
  try:
    fastaFileBasename = os.path.basename(fastaFile)
    os.symlink(fastaFile, fastaFileBasename)

    cruxIndex = '.'.join(fastaFile.split('.')[0:-1]) + "_crux_index"
    cruxIndexBasename = os.path.basename(cruxIndex)
    os.symlink(cruxIndex, cruxIndexBasename)

    ms2Basename = os.path.basename(ms2File)
    os.symlink(ms2File, ms2Basename)

    tmpSpectrum = "temp.ms2"
    runner = Crux.GetSpectrumRunner(scan, ms2Basename, tmpSpectrum)
    runner.run()
    spectrumColl = SpectrumCollection.Ms2SpectrumCollection(tmpSpectrum)
    spectrum = spectrumColl.allSpectra()[0]
    spectrum["charge"] = charge
  except Exception, val:
    raise Exception, "Failed to get spectrum: %s" % str(val)

  # search the spectrum
  try:
    parameterFile = "%s/models/crux.params" % os.getenv("RIPTIDE_HOME")
    searchRunner = Crux.SearchSpectraRunner(tmpSpectrum, fastaFileBasename, parameterFile=parameterFile)
    searchRunner.spectrum_charge = charge 
    searchRunner.run()
    filteredPeps = searchRunner.getPeptides(maxPeptides=500)
  except Exception, val: 
    raise Exception, "Failed to get peptides for spectrum: %s" % str(val)


  try:
    psmCollection = MatchCollection.PsmMatchCollection()
    psmCollection.addPsmsForSpectrum(spectrum, filteredPeps)
  except Exception, val: 
    raise Exception, "Failed to prepare spectrum: %s" % str(val)

  singleModel = GMTKmodel.gmtkModelFactory(SINGLE_MODEL_NAME, charge=charge, training=False, paramPrefix=paramPrefix)
  pairModel = GMTKmodel.gmtkModelFactory(PAIR_MODEL_NAME, charge=charge, paramPrefix=paramPrefix)

  try:
    psmCollection["file"] = "testPsmColl" # the routines below need this
    print >>sys.stderr, "Generating single scores"
    singleScores = sionLor(psmCollection, singleModel, paramPrefix)
    print >>sys.stderr, "Generating pair scores"
    pairScores = sionLor(psmCollection, pairModel, pairedParamPrefix)
  except Exception, val: 
    raise Exception, "Failed to process spectrum: %s" % str(val)

  singleSvmInputFile   = spectrumUid + "-single." + "svminput"
  pairSvmInputFile     = spectrumUid + "-pair." + "svminput"
  mergedSvmInputFile   = spectrumUid + "-merged." + "svminput"

  try: 
    fh = open(singleSvmInputFile, 'w'); fh.write(singleScores.svmString()); fh.close();
    fh = open(pairSvmInputFile, 'w'); fh.write(pairScores.svmString()); fh.close();
    FixSvm.merge(pairSvmInputFile, singleSvmInputFile, mergedSvmInputFile)
    
  except Exception, e: 
    raise Exception, "Could not output svminput %s" % str(e)

  # START
  try:
    svmFullPrefix = originalDir + os.sep + paramPrefix
    pep2score = getPepToScore(mergedSvmInputFile, svmFullPrefix)
  except Exception, e:
    raise Exception, "Could not run svm %s" % str(e)

  # output scores
  try:
    scoreCount = {}
    maxScore = max(pep2score.values())
    for pep in pep2score.keys(): 
      score = pep2score[pep]
      if score == maxScore: topPep = pep 
      if scoreCount.has_key(score): scoreCount[score] += 1
      else: scoreCount[score] = 1

    maxCount = max(scoreCount.values())
    if maxCount * 20 > len(pep2score.keys()):
      topScore = -10.0
      topPep = DUPSCORE
    else:
      topScore = GaussianMixture.transformTopScore(pep2score.values())
    svmFh = open(originalDir + os.sep + mergedSvmInputFile, "w")
    readFh = open(mergedSvmInputFile)
    for line in readFh:
      if line.rfind(topPep) != -1:
        svmFh.write(line)
        break
    svmFh.close()
    readFh.close()
    fh = open(originalDir + os.sep + spectrumUid + "." + "scores", 'w')
    fh.write("%s\t%s\n" % (topPep, topScore))
    fh.close()


  except Exception, e: 
    raise Exception, "Could not transform or output scores! %s" % str(e)

  try:

    # shutil.copy(mergedSvmInputFile, originalDir + os.sep + ) 
    fh = open(mergedSvmInputFile)
    fhOut = open(originalDir + os.sep + mergedSvmInputFile, "w") 
    for line in fh:
      if line.find(topPep) != -1:
        fhOut.write(line)
    fh.close()
    fhOut.close()
    cruxMatchFiles = filter(\
        lambda x: x.startswith("crux_match"), os.listdir(os.getcwd()))
    for file in cruxMatchFiles:
      shutil.copy(file, originalDir + os.sep + file) 
  except Exception, e:
    raise Exception, "Could not copy svminput or crux_match files! %s" % str(e)

  os.chdir(WORKING_DIR)
  if CLEANUP == True:
    try: shutil.rmtree(TEMP_DIR)
    except: raise Exception, "Failed to remove TEMP_DIR %s!" % TEMP_DIR

#-------------------------------------------------------------------------
if __name__ == "__main__":
  signal.signal(signal.SIGTERM, sigtermHandler)
  try: main(sys.argv)
  except Exception, e: abort(str(e))

  print >>sys.stderr, "SUCCESS - %s %s %s %s" % \
    (sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])

  sys.exit(0)
