#---------------------------------------------------------------------------
# Name: generatePeptides.py
#---------------------------------------------------------------------------
# Author: $Author: aklammer $
#---------------------------------------------------------------------------
# Revision: $Revision: 1.4 $
#---------------------------------------------------------------------------
# Inputs:
#  - a <prefix>.fasta file from which to extract aa frequencies
#  - a pickled SpectrumCollection with spectra 
#    for which to generate candidate and null peptides 
#--------------------------------------------------------------------------
# Outputs:
#  - a directory <prefix>.pepdir containing two files for each spectrum:
#      - spectrum-id.null - synthetic peptides for the null distribution 
#      - spectrum-id.cand - candidate peptides for matches
#  - an output file <prefix>.peps mapping the spectra to their peptide files: 
# 
#     spectrum id \t null peps file \t cand peps file \n
#---------------------------------------------------------------------------
# NOTE: This is meant to faithfully represent PepHMM. For a real database
# search, you would want to have peptides near the same m/z (I think) not the
# same mass, since it is the m/z that is known. In PepHMM, we deal only with
# +2 spectra, so this is not an issue. 
#---------------------------------------------------------------------------
# MIN_MASS = 400
MIN_MASS = 50
# MAX_MASS = 5000
MAX_MASS = 5000
# MIN_PEPTIDES = 1
MIN_PEPTIDES = 20
MIN_PEPTIDE_LEN = 3 
ITERATIONS = 4
MASS_WINDOW = 2
MAX_PEPTIDES = 500


#---------------------------------------------------------------------------
import os, sys
if __name__ == "__main__":
  if len(sys.argv) < 2:
    print >>sys.stderr, """
Usage: %s <seq db prefix> <pickled object> 
    <seq db> - A <prefix>.fasta file for generating null and candidate peptides
    <pickled object> 
             - The pickled SpectrumCollection (with suffix .pickledscoll) or 
               PsmMatchCollection (with suffix .pickledpsm) object that 
               will be searched.
             
""" % os.path.basename(sys.argv[0])
    sys.exit(1)

#---------------------------------------------------------------------------
try:
  import Bio.Fasta
except:
  print >>sys.stderr, "Could not import Bio.Fasta"

import random
import Mass
import SpectrumCollection, Spectrum
import cPickle

#---------------------------------------------------------------------------
def frequencyToRandDict(freqDict):
  """
  return a list of length PRECISION where values are the keys of input dict.
  Each key in input dict is present as a value in output dict proportional
  to its corresponding value in input freqDict.
  """

  idx = 0
  rand = []
  for key in freqDict:
    for subIdx in range(int(freqDict[key]*PRECISION)):
      rand.append(key)
      idx+=1
  if len(rand) < PRECISION: rand.append(rand[-1])
  return rand


def reverse(peptide):
  rev = ""
  for i in range(len(peptide)):
    rev+=peptide[len(peptide)-1-i]
  return rev

#--------------------------------------------------------------------
#--------------------------------------------------------------------
# generate a random tryptic 
# peptide with aa freq aaFreq within one of the masses in massesAllowed
#--------------------------------------------------------------------
def generateTrypticPeptide(randDict, kFreq=0.5, rFreq=0.5, massesAllowed=None, z=2):
  """
i aa:frequency 
i mass:exist
o peptide
  """
  kThresh = kFreq / (kFreq + rFreq)
   
  while(True):
    endingAA = ""
    if random.random() < kThresh: endingAA = 'K'
    else: endingAA = 'R'

    peptide = endingAA
    mass = Mass.peptide_mass(endingAA, z=2)

    for i in range(ITERATIONS):

      while mass < MAX_MASS:
        raa = randDict[ int(len(randDict)*random.random()) ] # Mass.randomAA()
        peptide += raa
        mass += Mass.mono_mass(raa)
        massInt = int(mass+0.5)
        if massesAllowed.has_key(massInt): 
          return reverse(peptide), massInt

      while MIN_MASS < mass and len(peptide) > 1:
        raa = peptide[-1]
        mass -= Mass.mono_mass(raa)
        peptide = peptide[0:len(peptide)-1] 

#-------------------------------------
def nullPeptideDict(fastaPrefix, massLeft):
  """
  return a dict mapping mass:[peptides] 
  uses only tryptic peptides
  """

  # count aa frequencies
  aaCount = {}
  for aa in Mass.CanonicalAminos: aaCount[aa] = 0
  totalCount = 0

  print >>sys.stderr, "Reading fasta file %s" % fastaPrefix
  fh = open(fastaPrefix + ".fasta")
  iterator = Bio.Fasta.Iterator(fh, Bio.Fasta.RecordParser())
  idx = 0
  while(True):
    idx += 1
    if idx % 10000 == 0: print >>sys.stderr, "Reached sequence %i" % idx
    curRecord = iterator.next()
    if curRecord is None: break
    seq = curRecord.sequence
    totalCount += len(seq)
    for idx in range(len(seq)): 
      if aaCount.has_key(seq[idx]): aaCount[seq[idx]]+=1

  fh.close()

  aaFrac = {} 
  for key in aaCount.keys(): aaFrac[key] = float(aaCount[key])/totalCount

  # rand dict maps int:aa
  
  # massLeft = {}
  # for i in range(MIN_MASS,MAX_MASS): massLeft[i] = True

  # maps mass:[peptide]
  mass2seq = {}
  
  # keep generating peptides until we have no masses left between MAX_MASS
  # and MIN_MASS
  rFreq = aaFrac.pop('R')
  kFreq = aaFrac.pop('K')

  randDict = frequencyToRandDict(aaFrac)

  while len(massLeft.keys()) > 0:
    peptide, massInt = generateTrypticPeptide(\
                    randDict, kFreq=kFreq, rFreq=rFreq, massesAllowed=massLeft)
    # if ( (mass > MAX_MASS) or (mass < MIN_MASS) ): continue

    if len(massLeft) % 100 == 0: 
      print >>sys.stderr, "Only %i masses left" % len(massLeft)
    # print "peptide = ", peptide
    # print "massLeft = ", len(massLeft)
    # print "-----"
    if not massLeft.has_key(massInt): continue

    # add to mass2seq
    if mass2seq.has_key(massInt): mass2seq[massInt].append(peptide) 
    else: mass2seq[massInt] = [ peptide ]

    # remove this mass if we have enough peptides
    if massLeft.has_key(massInt):
      if len(mass2seq[massInt]) >= MIN_PEPTIDES: 
        massLeft.pop(massInt)

  return mass2seq

#-----------------------------------------
# MAIN
#-----------------------------------------
if __name__ == "__main__":

  fastaPrefix = sys.argv[1]
  objPrefix = sys.argv[2].split('.')[0]
  
  # all masses are real masses. that is, masses of the charged peptide
  suid2mass = {}
  print >>sys.stderr, "Loading PsmCollection %s" % sys.argv[2]
  psmColl = cPickle.load(open(sys.argv[2]))
  for spectrum in psmColl.allSpectra():
    suid2mass[spectrum.uid()] = int(float(spectrum["assumed m"]) + Mass.mono_mass('H_plus') + 0.5)

  massesAllowed = {}
  for value in suid2mass.values(): 
    for mass in range(value-MASS_WINDOW, value + MASS_WINDOW + 1):
      massesAllowed[mass] = 1

  print >>sys.stderr, "Creating null peptides for %s" % fastaPrefix
  nullDict = nullPeptideDict(fastaPrefix, massesAllowed)
  nullPickle = open(fastaPrefix + ".nulldict", "w")
  cPickle.dump(nullDict, nullPickle)
  nullPickle.close()

  for match in psmColl["matches"]:
    suid = match.spectrum.uid()
    nullPeps = nullDict[suid2mass[suid]]
    randompep = nullPeps[ int( random.random() * len(nullPeps) )]
    print randompep
    match.peptide = randompep

  randomObs = objPrefix  + ",trurand.pickledpsm"
  cPickle.dump(psmColl, open(randomObs, "w"))

